import numpy as np
from ccalafiore.array import samples_in_arr1_are_in_arr2  # , advanced_indexing
from ccalafiore.combinations import (
    trials_to_conditions, conditions_to_n_conditions, n_conditions_to_combinations, conditions_to_combinations)
from ccalafiore.preprocessing.axes_to_variables_table import fast_balanced as cc_a2v


def from_2_arrays(
        array_variables_staying,
        array_variables_adding_axes,
        axis_samples,
        axis_variables_table,
        axes_inserting=None,
        balance=True,
        dtype=None,
        format_and_check=True):

    axis_variables_table_input = axis_variables_table
    axis_samples_input = axis_samples

    shape_array_variables_staying = np.asarray(array_variables_staying.shape)
    n_axes_array_input = shape_array_variables_staying.size

    if balance:

        n_variables_table_staying = shape_array_variables_staying[axis_variables_table_input]

        if format_and_check:
            n_axes_array_variables_staying = n_axes_array_input

            shape_array_variables_adding_axes = np.asarray(array_variables_adding_axes.shape)
            n_axes_array_variables_adding_axes = shape_array_variables_adding_axes.size

            # check point 1
            if n_axes_array_variables_staying != n_axes_array_variables_adding_axes:
                raise ValueError('dimension mismatch')

            if axis_variables_table_input < 0:
                axis_variables_table_input += n_axes_array_input
            if axis_samples_input < 0:
                axis_samples_input += n_axes_array_input

            # check point 5
            if axis_samples_input == axis_variables_table_input:
                raise ValueError('axis_samples_input and axis_variables_table_input must be different')

            # check point 2
            indexes_logical = np.arange(n_axes_array_input) != axis_variables_table_input
            if np.any(shape_array_variables_staying[indexes_logical] !=
                      shape_array_variables_adding_axes[indexes_logical]):
                raise ValueError('dimension mismatch')

            n_variables_table_adding_axes = shape_array_variables_adding_axes[axis_variables_table_input]

            # format axes_inserting
            n_axes_inserting = n_variables_table_adding_axes
            n_axes_array_output = n_axes_array_input + n_axes_inserting
            try:
                len(axes_inserting)
                axes_inserting = np.asarray(axes_inserting, dtype=int)
                axes_inserting[axes_inserting < 0] += n_axes_array_output
                # check point 3
                if np.sum(axes_inserting[0] == axes_inserting) > 1:
                    raise ValueError('axes_inserting cannot contain repeated values')
            except TypeError:
                if axes_inserting is None:
                    axes_inserting = np.arange(n_axes_inserting)
                else:
                    if axes_inserting < 0:
                        axes_inserting += n_axes_array_output
                    axes_inserting = np.asarray([axes_inserting], dtype=int)

        else:
            n_axes_inserting = axes_inserting.size
            n_axes_array_output = n_axes_array_input + n_axes_inserting

        axes_array_input = np.arange(n_axes_array_input)
        axes_non_axis_variables_table_input = axes_array_input[axes_array_input != axis_variables_table_input]
        axes_other_array_input = axes_non_axis_variables_table_input[
            axes_non_axis_variables_table_input != axis_samples_input]
        n_axes_other_array_input = axes_other_array_input.size

        indexes_array_variables_staying = np.empty(n_axes_array_input, dtype=object)
        indexes_array_variables_staying[:] = slice(None)

        axis_variables_in_combinations = int(axis_variables_table_input > axis_samples_input)
        axis_combinations_in_combinations = int(not (bool(axis_variables_in_combinations)))

        if n_axes_other_array_input == 0:
            conditions_variables_table_adding_axes = trials_to_conditions(
                array_variables_adding_axes, axis_combinations=axis_combinations_in_combinations)
        elif n_axes_other_array_input > 0:
            indexes_array_variables_adding_axes = np.copy(indexes_array_variables_staying)
            indexes_array_variables_adding_axes[axes_other_array_input] = 0
            array_variables_adding_axes_1 = array_variables_adding_axes[tuple(indexes_array_variables_adding_axes)]
            conditions_variables_table_adding_axes = trials_to_conditions(
                array_variables_adding_axes_1, axis_combinations=axis_combinations_in_combinations)

        n_conditions_variables_table_adding_axes = conditions_to_n_conditions(
            conditions_variables_table_adding_axes)
        n_combinations_variables_table_adding_axes = int(np.prod(
            n_conditions_variables_table_adding_axes))

        combinations_variables_table_adding_axes = conditions_to_combinations(
            conditions_variables_table_adding_axes,
            axis_combinations=axis_combinations_in_combinations)

        combinations_axes_inserting = n_conditions_to_combinations(
            n_conditions_variables_table_adding_axes)

        axis_variables_table_output = axis_variables_table_input + np.sum(
            axes_inserting <= axis_variables_table_input)
        axis_samples_output = axis_samples_input + np.sum(axes_inserting <= axis_samples_input)
        changed = True
        while changed:
            changed = False
            while axis_variables_table_output in axes_inserting:
                axis_variables_table_output += 1
                changed = True
            while axis_samples_output in axes_inserting:
                axis_samples_output += 1
                changed = True
            if axis_variables_table_output == axis_samples_output:
                changed = True
                if axis_variables_table_input > axis_samples_input:
                    axis_variables_table_output += 1
                elif axis_variables_table_input < axis_samples_input:
                    axis_samples_output += 1

        axes_array_output = np.arange(n_axes_array_output)
        axes_non_axis_variables_table_output = axes_array_output[axes_array_output != axis_variables_table_output]
        axes_other_output = axes_non_axis_variables_table_output[
            axes_non_axis_variables_table_output != axis_samples_output]
        axes_other_output = axes_other_output[np.logical_not(
            samples_in_arr1_are_in_arr2(axes_other_output, axes_inserting))]
        # axes_inserting_sorted = np.sort(axes_inserting)

        shape_array_output = np.empty(n_axes_array_output, dtype=int)
        shape_array_output[axes_inserting] = n_conditions_variables_table_adding_axes
        shape_array_output[axis_variables_table_output] = n_variables_table_staying
        shape_array_output[axis_samples_output] = (
                shape_array_variables_staying[axis_samples_input] / n_combinations_variables_table_adding_axes)
        shape_array_output[axes_other_output] = shape_array_variables_staying[axes_other_array_input]
        if dtype is None:
            dtype = array_variables_staying.dtype
        array_output = np.empty(shape_array_output, dtype=dtype)

        indexes_output = np.empty(n_axes_array_output, dtype=object)
        indexes_output[:] = slice(None)

        indexes_combinations = np.empty(2, dtype=object)
        indexes_combinations[axis_variables_in_combinations] = slice(None)

        if n_axes_other_array_input == 0:
            for i in range(n_combinations_variables_table_adding_axes):
                indexes_output[axes_inserting] = combinations_axes_inserting[i]
                indexes_combinations[axis_combinations_in_combinations] = slice(i, i + 1)
                indexes_array_variables_staying[axis_samples_input] = np.all(
                    array_variables_adding_axes ==
                    combinations_variables_table_adding_axes[tuple(indexes_combinations)],
                    axis=axis_variables_in_combinations)
                array_output[tuple(indexes_output)] = array_variables_staying[tuple(
                    indexes_array_variables_staying)]

        elif n_axes_other_array_input > 0:
            indexes_array_variables_staying_j = np.empty(2, dtype=object)
            indexes_array_variables_staying_j[axis_variables_in_combinations] = slice(None)

            combinations_axes_other = n_conditions_to_combinations(
                shape_array_variables_staying[axes_other_array_input])
            n_combinations_axes_other = combinations_axes_other.shape[0]
            for j in range(n_combinations_axes_other):
                combinations_axes_other_j = combinations_axes_other[j]
                indexes_output[axes_other_output] = combinations_axes_other_j
                indexes_array_variables_staying[axes_other_array_input] = combinations_axes_other_j
                indexes_array_variables_adding_axes[axes_other_array_input] = combinations_axes_other_j
                array_variables_adding_axes_j = array_variables_adding_axes[
                    tuple(indexes_array_variables_adding_axes)]
                array_variables_staying_j = array_variables_staying[tuple(
                    indexes_array_variables_staying)]
                for i in range(n_combinations_variables_table_adding_axes):
                    indexes_output[axes_inserting] = combinations_axes_inserting[i]
                    indexes_combinations[axis_combinations_in_combinations] = slice(i, i + 1)
                    indexes_array_variables_staying_j[axis_combinations_in_combinations] = np.all(
                        array_variables_adding_axes_j ==
                        combinations_variables_table_adding_axes[tuple(indexes_combinations)],
                        axis=axis_variables_in_combinations)
                    array_output[tuple(indexes_output)] = array_variables_staying_j[tuple(
                        indexes_array_variables_staying_j)]

    else:
        # n_variables_table_staying = shape_array_variables_staying[axis_variables_table_input]

        if format_and_check:
            n_axes_array_variables_staying = n_axes_array_input

            shape_array_variables_adding_axes = np.asarray(array_variables_adding_axes.shape)
            n_axes_array_variables_adding_axes = shape_array_variables_adding_axes.size

            # check point 1
            if n_axes_array_variables_staying != n_axes_array_variables_adding_axes:
                raise ValueError('dimension mismatch')

            if axis_variables_table_input < 0:
                axis_variables_table_input += n_axes_array_input
            if axis_samples_input < 0:
                axis_samples_input += n_axes_array_input

            # check point 5
            if axis_samples_input == axis_variables_table_input:
                raise ValueError('axis_samples_input and axis_variables_table_input must be different')

            # check point 2
            indexes_logical = np.arange(n_axes_array_input) != axis_variables_table_input
            if np.any(shape_array_variables_staying[indexes_logical] !=
                      shape_array_variables_adding_axes[indexes_logical]):
                raise ValueError('dimension mismatch')

            n_variables_table_adding_axes = shape_array_variables_adding_axes[axis_variables_table_input]

            # format axes_inserting
            n_axes_inserting = n_variables_table_adding_axes
            n_axes_array_output_object = n_axes_inserting
            try:
                len(axes_inserting)
                axes_inserting = np.asarray(axes_inserting, dtype=int)
                axes_inserting[axes_inserting < 0] += n_axes_array_output_object
                # check point 3
                if np.sum(axes_inserting[0] == axes_inserting) > 1:
                    raise ValueError('axes_inserting cannot contain repeated values')

            except TypeError:
                if axes_inserting is None:
                    axes_inserting = np.arange(n_axes_inserting)
                else:
                    if axes_inserting < 0:
                        axes_inserting += n_axes_array_output_object
                    axes_inserting = np.asarray([axes_inserting], dtype=int)

            if dtype is not None:
                array_variables_staying = array_variables_staying.astype(dtype)

        else:
            n_axes_inserting = axes_inserting.size
            n_axes_array_output_object = n_axes_inserting

        axes_array_input = np.arange(n_axes_array_input)
        axes_non_axis_variables_table_input = axes_array_input[axes_array_input != axis_variables_table_input]
        axes_other_array_input = axes_non_axis_variables_table_input[
            axes_non_axis_variables_table_input != axis_samples_input]
        n_axes_other_array_input = axes_other_array_input.size

        indexes_array_variables_staying = np.empty(n_axes_array_input, dtype=object)
        indexes_array_variables_staying[:] = slice(None)

        axis_variables_in_combinations = int(axis_variables_table_input > axis_samples_input)
        axis_combinations_in_combinations = int(not (bool(axis_variables_in_combinations)))

        if n_axes_other_array_input == 0:
            array_variables_adding_axes_object = array_variables_adding_axes
            conditions_variables_table_adding_axes_object = trials_to_conditions(
                array_variables_adding_axes_object, axis_combinations=axis_combinations_in_combinations)
        elif n_axes_other_array_input > 0:
            indexes_object = np.empty(2, dtype=object)
            indexes_object[axis_combinations_in_combinations] = slice(None)
            indexes_object[axis_variables_in_combinations] = slice(
                n_axes_other_array_input, n_axes_other_array_input + 2)
            indexes_not_object = np.copy(indexes_object)
            indexes_not_object[axis_variables_in_combinations] = slice(n_axes_other_array_input)

            array_variables_adding_axes = cc_a2v(
                array_variables_adding_axes,
                axes_removing_input=axes_other_array_input,
                axis_samples_input=axis_samples_input,
                axis_variables_table_input=axis_variables_table_input,
                variables_inserting_table_output=np.arange(n_axes_other_array_input),
                variables_staying_table_input=None,
                dtype=None)
            array_variables_adding_axes_object = array_variables_adding_axes[tuple(indexes_object)]
            array_variables_adding_axes_not_object = array_variables_adding_axes[tuple(indexes_not_object)]

            array_variables_staying = cc_a2v(
                array_variables_staying,
                axes_removing_input=axes_other_array_input,
                axis_samples_input=axis_samples_input,
                axis_variables_table_input=axis_variables_table_input,
                variables_inserting_table_output=np.arange(n_axes_other_array_input),
                variables_staying_table_input=None,
                dtype=None)[tuple(indexes)]

            conditions_variables_table_adding_axes_object = trials_to_conditions(
                array_variables_adding_axes_object, axis_combinations=axis_combinations_in_combinations)

            n_conditions_variables_table_adding_axes_not_object = shape_array_variables_adding_axes[
                axes_other_array_input]

            combinations_axes_inserting_not_object = n_conditions_to_combinations(
                n_conditions_variables_table_adding_axes_not_object)

        n_conditions_variables_table_adding_axes_object = conditions_to_n_conditions(
            conditions_variables_table_adding_axes_object)
        n_combinations_variables_table_adding_axes_object = int(np.prod(
            n_conditions_variables_table_adding_axes_object))

        combinations_variables_table_adding_axes_object = conditions_to_combinations(
            conditions_variables_table_adding_axes_object,
            axis_combinations=axis_combinations_in_combinations)

        combinations_axes_inserting_object = n_conditions_to_combinations(
            n_conditions_variables_table_adding_axes_object)

        # axes_array_output_object = np.arange(n_axes_array_output_object)
        shape_array_output_object = np.empty(n_axes_array_output_object, dtype=int)
        shape_array_output_object[axes_inserting] = n_conditions_variables_table_adding_axes
        array_output = np.empty(shape_array_output_object, dtype=object)

        indexes_output_object = np.empty(n_axes_array_output_object, dtype=object)

        indexes_combinations = np.empty(2, dtype=object)
        indexes_combinations[axis_variables_in_combinations] = slice(None)

        # if n_axes_other_array_input == 0:
        for i in range(n_combinations_variables_table_adding_axes_object):
            indexes_output_object[axes_inserting] = combinations_axes_inserting_object[i]
            indexes_combinations[axis_combinations_in_combinations] = slice(i, i + 1)
            indexes_array_variables_staying[axis_samples_input] = np.all(
                array_variables_adding_axes_object ==
                combinations_variables_table_adding_axes_object[tuple(indexes_combinations)],
                axis=axis_variables_in_combinations)
            array_output[tuple(indexes_output_object)] = array_variables_staying[tuple(
                indexes_array_variables_staying)]

        if n_axes_other_array_input > 0:

            array_output[tuple(indexes_output_object)] = array_variables_staying[tuple(
                indexes_array_variables_staying)]

        # elif n_axes_other_array_input > 0:
        #
        #     shape_array_output = np.copy(shape_array_variables_staying)
        #     shape_array_output[axis_samples_input] = 0 # (
        #             # shape_array_variables_staying[axis_samples_input] / n_combinations_variables_table_adding_axes)
        #
        #     dtype = array_variables_staying.dtype
        #     array_output_tmp = np.empty(shape_array_output, dtype=dtype)
        #
        #     for i in range(n_combinations_variables_table_adding_axes):
        #         indexes_output_object[axes_inserting] = combinations_axes_inserting[i]
        #         array_output[tuple(indexes_output_object)] = np.copy(array_output_tmp)
        #
        #     indexes_output = np.empty(n_axes_array_input, dtype=object)
        #     indexes_output[:] = slice(None)
        #
        #     indexes_array_variables_staying_j = np.empty(2, dtype=object)
        #     indexes_array_variables_staying_j[axis_variables_in_combinations] = slice(None)
        #
        #     indexes_array_variables_adding_axes = np.copy(indexes_array_variables_staying)
        #
        #     combinations_axes_other = n_conditions_to_combinations(
        #         shape_array_variables_staying[axes_other_array_input])
        #     n_combinations_axes_other = combinations_axes_other.shape[0]
        #     for j in range(n_combinations_axes_other):
        #         combinations_axes_other_j = combinations_axes_other[j]
        #         indexes_output[axes_other_array_input] = combinations_axes_other_j
        #         indexes_output_tuple = tuple(indexes_output)
        #         indexes_array_variables_staying[axes_other_array_input] = combinations_axes_other_j
        #         indexes_array_variables_adding_axes[axes_other_array_input] = combinations_axes_other_j
        #         array_variables_adding_axes_j = array_variables_adding_axes[
        #             tuple(indexes_array_variables_adding_axes)]
        #         array_variables_staying_j = array_variables_staying[tuple(
        #             indexes_array_variables_staying)]
        #         for i in range(n_combinations_variables_table_adding_axes):
        #             indexes_output_object[axes_inserting] = combinations_axes_inserting[i]
        #             indexes_combinations[axis_combinations_in_combinations] = slice(i, i + 1)
        #             indexes_array_variables_staying_j[axis_combinations_in_combinations] = np.all(
        #                 array_variables_adding_axes_j ==
        #                 combinations_variables_table_adding_axes[tuple(indexes_combinations)],
        #                 axis=axis_variables_in_combinations)
        #
        #             array_output[tuple(indexes_output_object)][indexes_output_tuple] = np.append(
        #                 array_output[tuple(indexes_output_object)][indexes_output_tuple],
        #                 array_variables_staying_j[tuple(indexes_array_variables_staying_j)],
        #                 axis=axis_samples_input)
                    # array_output[tuple(indexes_output_object)][indexes_output_tuple] = array_variables_staying_j[
                    #     tuple(indexes_array_variables_staying_j)]

    return array_output


def from_2_arrays_advanced(
        array_variables_staying,
        array_variables_adding_axes,
        axis_samples,
        axis_variables_table,
        variables_table_adding_axes=None,
        axes_inserting=None,
        variables_table_staying=None,
        balance=True,
        dtype=None,
        format_and_check=True):

    # Notes:
    # 1) uses the efficient function slice for indexes;
    # 2) it splits the array in two arrays first: array_variables_adding_axes and array_variables_staying.
    #    Then, it makes the array output from those two arrays;
    # 3) it assumes that the numbers of samples (or trials) in the axis_samples for all combinations
    #    of variables_table_adding_axes are equal;
    # 4) it does not assumes that the order of samples (or trials) in the axis_samples are is the same
    #    in each combination of variables' axes. The order is based on the order of the combinations
    #    of variables' conditions in axis_variables_table. In other words, the samples can be in random order
    #    (or in different order) in each combinations of variables axes/conditions.

    # Input requirements:
    # 1) the numbers of axes of the two arrays must be equal;
    # 2) the shapes of the 2 arrays must be equal, except for the axis_variable_table,
    #    i.e. the axis that contains the variables of the tables;
    # 3) axes_inserting cannot contain repeated values;
    # 4) variables_table_adding_axes cannot contain repeated values;
    # 5) shapes of axes_inserting and variables_table_adding_axes must be equal
    # 6) axis_samples != axis_variables_table
    # 7) the numbers of samples (or trials) in the axis_samples for all combinations
    #    of variables_table_adding_axes are equal

    # Import requirements:
    # 1) import numpy as np
    # 2) from ccalafiore.combinations import n_conditions_to_combinations
    # 3) from ccalafiore.combinations import trials_to_conditions
    # 4) from ccalafiore.combinations import conditions_to_n_conditions
    # 5) from ccalafiore.combinations import conditions_to_combinations
    # 6) from ccalafiore.array import samples_in_arr1_are_in_arr2

    # from ccalafiore.combinations import \
    #     trials_to_conditions, conditions_to_n_conditions, n_conditions_to_combinations, conditions_to_combinations
    # from ccalafiore.array import samples_in_arr1_are_in_arr2

    axis_variables_table_input = axis_variables_table
    axis_samples_input = axis_samples

    shape_array_variables_staying = np.asarray(array_variables_staying.shape)
    n_axes_array_variables_staying = shape_array_variables_staying.size
    n_variables_table_in_array_variables_staying = shape_array_variables_staying[axis_variables_table_input]
    variables_table_in_array_variables_staying = np.arange(n_variables_table_in_array_variables_staying)

    shape_array_variables_adding_axes = np.asarray(array_variables_adding_axes.shape)
    n_axes_array_variables_adding_axes = shape_array_variables_adding_axes.size
    n_variables_table_in_array_variables_adding_axes = shape_array_variables_adding_axes[axis_variables_table_input]
    variables_table_in_array_variables_adding_axes = np.arange(n_variables_table_in_array_variables_adding_axes)

    if format_and_check:
        # check point 1
        if n_axes_array_variables_staying != n_axes_array_variables_adding_axes:
            raise ValueError('dimension mismatch')
        else:
            n_axes_array_input = n_axes_array_variables_staying

        if axis_variables_table_input < 0:
            axis_variables_table_input += n_axes_array_input
        if axis_samples_input < 0:
            axis_samples_input += n_axes_array_input

        # check point 6
        if axis_samples_input == axis_variables_table_input:
            raise ValueError('axis_samples_input and axis_variables_table_input must be different')

        # check point 2
        indexes_logical = np.arange(n_axes_array_input) != axis_variables_table_input
        if np.any(shape_array_variables_staying[indexes_logical] != shape_array_variables_adding_axes[indexes_logical]):
            raise ValueError('dimension mismatch')

        n_variables_table_in_array_variables_staying = shape_array_variables_staying[axis_variables_table_input]
        variables_table_in_array_variables_staying = np.arange(n_variables_table_in_array_variables_staying)

        n_variables_table_in_array_variables_adding_axes = shape_array_variables_adding_axes[axis_variables_table_input]
        variables_table_in_array_variables_adding_axes = np.arange(n_variables_table_in_array_variables_adding_axes)

        # format variables_table_adding_axes
        try:
            n_variables_table_adding_axes = len(variables_table_adding_axes)
            variables_table_adding_axes = np.asarray(variables_table_adding_axes, dtype=int)
            variables_table_adding_axes[variables_table_adding_axes < 0] += \
                n_variables_table_in_array_variables_adding_axes
            # check point 4
            if np.sum(variables_table_adding_axes[0] == variables_table_adding_axes) > 1:
                raise ValueError('variables_table_adding_axes cannot contain repeated values')
        except TypeError:
            if variables_table_adding_axes is None:
                variables_table_adding_axes = np.arange(n_variables_table_in_array_variables_adding_axes)
                n_variables_table_adding_axes = n_variables_table_in_array_variables_adding_axes
            else:
                if variables_table_adding_axes < 0:
                    variables_table_adding_axes += n_variables_table_in_array_variables_adding_axes
                variables_table_adding_axes = np.asarray([variables_table_adding_axes], dtype=int)
                n_variables_table_adding_axes = 1

        # format variables_table_staying
        try:
            n_variables_table_staying = len(variables_table_staying)
            variables_table_staying = np.asarray(variables_table_staying, dtype=int)
            variables_table_staying[variables_table_staying < 0] += n_variables_table_in_array_variables_staying
        except TypeError:
            if variables_table_staying is None:
                variables_table_staying = np.arange(n_variables_table_in_array_variables_staying)
                n_variables_table_staying = n_variables_table_in_array_variables_staying
            else:
                if variables_table_staying < 0:
                    variables_table_staying += n_variables_table_in_array_variables_staying
                variables_table_staying = np.asarray([variables_table_staying], dtype=int)
                n_variables_table_staying = 1

        # format axes_inserting
        n_axes_inserting = n_variables_table_adding_axes
        if balance:
            n_axes_array_output = n_axes_array_input + n_axes_inserting
        else:
            n_axes_array_output = n_axes_inserting
        try:
            n_axes_inserting = len(axes_inserting)
            axes_inserting = np.asarray(axes_inserting, dtype=int)
            axes_inserting[axes_inserting < 0] += n_axes_array_output
            # check point 3
            if np.sum(axes_inserting[0] == axes_inserting) > 1:
                raise ValueError('axes_inserting cannot contain repeated values')
            # check point 5
            if n_variables_table_adding_axes != n_axes_inserting:
                raise ValueError('Shapes of axes_inserting and variables_table_adding_axes must be equal')
        except TypeError:
            if axes_inserting is None:
                axes_inserting = np.arange(n_axes_inserting)
            else:
                if axes_inserting < 0:
                    axes_inserting += n_axes_array_output
                axes_inserting = np.asarray([axes_inserting], dtype=int)
                n_axes_inserting = 1
                # check point 5
                if n_variables_table_adding_axes != n_axes_inserting:
                    raise ValueError('Shapes of axes_inserting and variables_table_adding_axes must be equal')
    else:
        n_variables_table_staying = variables_table_staying.size
        n_variables_table_adding_axes = variables_table_adding_axes.size

    indexes = None
    if ((n_variables_table_staying != n_variables_table_in_array_variables_staying) or
            np.any(variables_table_staying != variables_table_in_array_variables_staying)):

        indexes = np.empty(n_axes_array_variables_staying, dtype=object)
        indexes[:] = slice(None)
        indexes[axis_variables_table_input] = variables_table_staying
        array_variables_staying = array_variables_staying[tuple(indexes)]

    if ((n_variables_table_adding_axes != n_variables_table_in_array_variables_adding_axes) or
            np.any(variables_table_adding_axes != variables_table_in_array_variables_adding_axes)):

        if indexes is None:
            indexes = np.empty(n_axes_array_variables_adding_axes, dtype=object)
            indexes[:] = slice(None)
        indexes[axis_variables_table_input] = variables_table_adding_axes
        array_variables_adding_axes = array_variables_adding_axes[tuple(indexes)]

    array_output = from_2_arrays(
        array_variables_staying,
        array_variables_adding_axes,
        axis_samples_input,
        axis_variables_table_input,
        axes_inserting=axes_inserting,
        balance=balance,
        dtype=dtype,
        format_and_check=False)

    return array_output


def from_1_array(
        array_input,
        axis_samples,
        axis_variables_table,
        variables_table_adding_axes,
        axes_inserting=None,
        variables_table_staying=None,
        balance=True,
        dtype=None,
        format_and_check=True):

    # Notes:
    # 1) uses the efficient function slice for indexes;
    # 2) it splits the array in two arrays first: array_variables_adding_axes and array_variables_staying.
    #    Then, it makes the array output from those two arrays;
    # 3) the numbers of samples (or trials) in the axis_samples for all combinations
    #    of variables_table_adding_axes have to be equal;
    # 4) it does not assumes that the order of samples (or trials) in the axis_samples are is the same
    #    in each combination of variables' axes. The order is based on the order of the combinations
    #    of variables' conditions in axis_variables_table. In other words, the samples can be in random order
    #    (or in different order) in each combinations of variables axes/conditions.

    # Input requirements:
    # 1) axes_inserting cannot contain repeated values;
    # 2) variables_table_adding_axes cannot contain repeated values;
    # 3) shapes of axes_inserting and variables_table_adding_axes must be equal
    # 4) axis_samples != axis_variables_table
    # 5) the numbers of samples (or trials) in the axis_samples for all combinations
    #    of variables_table_adding_axes are equal

    # Import requirements:
    # 1) import numpy as np
    # 2) from ccalafiore.combinations import n_conditions_to_combinations
    # 3) from ccalafiore.combinations import trials_to_conditions
    # 4) from ccalafiore.combinations import conditions_to_n_conditions
    # 5) from ccalafiore.combinations import conditions_to_combinations
    # 6) from ccalafiore.array import samples_in_arr1_are_in_arr2

    # from ccalafiore.combinations import \
    #     trials_to_conditions, conditions_to_n_conditions, n_conditions_to_combinations, conditions_to_combinations
    # from ccalafiore.array import samples_in_arr1_are_in_arr2

    axis_variables_table_input = axis_variables_table
    axis_samples_input = axis_samples

    shape_array_input = np.asarray(array_input.shape)
    n_axes_array_input = shape_array_input.size
    n_variables_table_input = shape_array_input[axis_variables_table_input]
    variables_table_input = np.arange(n_variables_table_input)

    if format_and_check:
        if axis_variables_table_input < 0:
            axis_variables_table_input += n_axes_array_input
        if axis_samples_input < 0:
            axis_samples_input += n_axes_array_input

        # check point 4
        if axis_samples_input == axis_variables_table_input:
            raise ValueError('axis_samples_input and axis_variables_table_input must be different')

        # format variables_table_adding_axes
        try:
            n_variables_table_adding_axes = len(variables_table_adding_axes)
            variables_table_adding_axes = np.asarray(variables_table_adding_axes, dtype=int)
            variables_table_adding_axes[variables_table_adding_axes < 0] += n_variables_table_input
            # check point 2
            if np.sum(variables_table_adding_axes[0] == variables_table_adding_axes) > 1:
                raise ValueError('variables_table_adding_axes cannot contain repeated values')
        except TypeError:
            if variables_table_adding_axes < 0:
                variables_table_adding_axes += n_variables_table_input
            variables_table_adding_axes = np.asarray([variables_table_adding_axes], dtype=int)
            n_variables_table_adding_axes = 1

        # format variables_table_staying
        try:
            n_variables_table_staying = len(variables_table_staying)
            variables_table_staying = np.asarray(variables_table_staying, dtype=int)
            variables_table_staying[variables_table_staying < 0] += n_variables_table_input
        except TypeError:
            if variables_table_staying is None:
                variables_table_staying = variables_table_input[np.logical_not(
                    samples_in_arr1_are_in_arr2(variables_table_input, variables_table_adding_axes))]
                n_variables_table_staying = variables_table_staying.size
            else:
                if variables_table_staying < 0:
                    variables_table_staying += n_variables_table_input
                variables_table_staying = np.asarray([variables_table_staying], dtype=int)
                n_variables_table_staying = 1

        # format axes_inserting
        n_axes_inserting = n_variables_table_adding_axes
        if balance:
            n_axes_array_output = n_axes_array_input + n_axes_inserting
        else:
            n_axes_array_output = n_axes_inserting
        try:
            n_axes_inserting = len(axes_inserting)
            axes_inserting = np.asarray(axes_inserting, dtype=int)
            axes_inserting[axes_inserting < 0] += n_axes_array_output
            # check point 1
            if np.sum(axes_inserting[0] == axes_inserting) > 1:
                raise ValueError('axes_inserting cannot contain repeated values')
            # check point 3
            if n_variables_table_adding_axes != n_axes_inserting:
                raise ValueError('Shapes of axes_inserting and variables_table_adding_axes must be equal')
        except TypeError:
            if axes_inserting is None:
                axes_inserting = np.arange(n_axes_inserting)
            else:
                if axes_inserting < 0:
                    axes_inserting += n_axes_array_output
                axes_inserting = np.asarray([axes_inserting], dtype=int)
                n_axes_inserting = 1
                # check point 3
                if n_variables_table_adding_axes != n_axes_inserting:
                    raise ValueError('Shapes of axes_inserting and variables_table_adding_axes must be equal')
    else:
        n_variables_table_staying = variables_table_staying.size
        n_variables_table_adding_axes = variables_table_adding_axes.size

    indexes = np.empty(n_axes_array_input, dtype=object)
    indexes[:] = slice(None)
    if ((n_variables_table_staying != n_variables_table_input) or
            np.any(variables_table_staying != variables_table_input)):
        indexes[axis_variables_table_input] = variables_table_staying
    array_variables_staying = array_input[tuple(indexes)]

    if ((n_variables_table_adding_axes != n_variables_table_input) or
            np.any(variables_table_adding_axes != variables_table_input)):
        indexes[axis_variables_table_input] = variables_table_adding_axes
    else:
        indexes[axis_variables_table_input] = slice(None)
    array_variables_adding_axes = array_input[tuple(indexes)]
    del array_input

    array_output = from_2_arrays(
        array_variables_staying,
        array_variables_adding_axes,
        axis_samples_input,
        axis_variables_table_input,
        axes_inserting=axes_inserting,
        balance=balance,
        dtype=dtype,
        format_and_check=False)

    return array_output



import numpy as np
from ccalafiore.array import samples_in_arr1_are_in_arr2, advanced_indexing
import ccalafiore.stats.basics as cc_stats_basics
from scipy.stats import t
# from scipy.stats.stats import _ttest_finish
from ccalafiore.combinations import n_conditions_to_combinations
# https://stackoverflow.com/questions/22611446/perform-2-sample-t-test


def scores_raw_to_scores_formatted(scores_raw, dtype=None):

    axis_object_independent_variable_raw = 0
    axis_object_comparisons_formatted = axis_object_independent_variable_raw
    axes_object_independent_variable_formatted = np.asarray([
        axis_object_comparisons_formatted + 1, axis_object_comparisons_formatted + 2], dtype=int)

    n_conditions_iv_per_comparisons = 2
    n_conditions_iv = scores_raw.size

    if dtype is not None:
        for c in range(n_conditions_iv):
            scores_raw[c] = scores_raw[c].astype(dtype, copy=False)

    n_axes_scores_formatted_object = 3
    shape_scores_formatted_object = np.empty(n_axes_scores_formatted_object, dtype=int)
    shape_scores_formatted_object[axis_object_comparisons_formatted] = n_conditions_iv_per_comparisons
    shape_scores_formatted_object[axes_object_independent_variable_formatted] = n_conditions_iv

    scores_formatted = np.empty(shape_scores_formatted_object, dtype=object)

    index_0_formatted_object = np.empty(n_axes_scores_formatted_object, dtype=object)
    index_0_formatted_object[:] = slice(None)
    index_1_formatted_object = np.copy(index_0_formatted_object)

    index_0_formatted_object[axis_object_comparisons_formatted] = 0
    index_1_formatted_object[axis_object_comparisons_formatted] = 1

    for i_condition in range(n_conditions_iv):

        scores_raw_0 = [scores_raw[i_condition]]

        index_0_formatted_object[axes_object_independent_variable_formatted[0]] = i_condition
        index_1_formatted_object[axes_object_independent_variable_formatted[1]] = i_condition

        scores_formatted[tuple(index_0_formatted_object)] = scores_raw_0
        scores_formatted[tuple(index_1_formatted_object)] = scores_raw_0
    return scores_formatted


def scores_formatted_to_df_in_unpaired_t_test(
        scores_formatted, axis_comparisons, axis_samples, equal_variances=False, keepdims=False, formt=True):

    if formt:
        shape_object_scores_formatted = np.asarray(scores_formatted.shape)
        n_axes_object_scores_formatted = shape_object_scores_formatted.size
        if axis_comparisons < 0:
            axis_comparisons += n_axes_object_scores_formatted

        indexes_object = np.empty(n_axes_object_scores_formatted, dtype=object)
        indexes_object[:] = 0
        shape_scores_formatted = np.asarray(scores_formatted[tuple(indexes_object)].shape)
        n_axes_scores_formatted = shape_scores_formatted.size
        if axis_samples < 0:
            axis_samples += n_axes_scores_formatted

    n_samples = cc_stats_basics.scores_to_n_samples(scores_formatted, axis_samples, keepdims=keepdims)

    shape_n_samples = np.asarray(n_samples.shape)
    n_axes_n_samples = shape_n_samples.size
    indexes_0 = np.empty(n_axes_n_samples, dtype=object)
    indexes_0[:] = slice(None)
    indexes_0[axis_comparisons] = 0

    indexes_1 = np.copy(indexes_0)
    indexes_1[axis_comparisons] = 1
    indexes_0_tuple = tuple(indexes_0)
    indexes_1_tuple = tuple(indexes_1)

    # EQUATIONS ARE AT https://www.statsdirect.co.uk/help/parametric_methods/utt.htm

    if equal_variances:
        df = n_samples[indexes_0_tuple] + n_samples[indexes_1_tuple] - 2
    else:
        variances = cc_stats_basics.scores_to_variances(scores_formatted, axis_samples, keepdims=keepdims)

        numerator_df = ((variances[indexes_0_tuple] / n_samples[indexes_0_tuple]) +
                        (variances[indexes_1_tuple] / n_samples[indexes_1_tuple]))**2

        denominator_df = ((((variances[indexes_0_tuple] / n_samples[indexes_0_tuple])**2) /
                           (n_samples[indexes_0_tuple] - 1)) +
                          (((variances[indexes_1_tuple] / n_samples[indexes_1_tuple]) ** 2) /
                           (n_samples[indexes_1_tuple] - 1)))

        df = numerator_df / denominator_df

    if keepdims:
        df = np.expand_dims(df, axis=axis_comparisons)

    return df


def scores_formatted_to_denominator_of_unpaired_t_test(
        scores_formatted, axis_comparisons, axis_samples, equal_variances=False, keepdims=False, formt=True):

    if formt:
        shape_object_scores_formatted = np.asarray(scores_formatted.shape)
        n_axes_object_scores_formatted = shape_object_scores_formatted.size
        if axis_comparisons < 0:
            axis_comparisons += n_axes_object_scores_formatted

        indexes_object = np.empty(n_axes_object_scores_formatted, dtype=object)
        indexes_object[:] = 0
        shape_scores_formatted = np.asarray(scores_formatted[tuple(indexes_object)].shape)
        n_axes_scores_formatted = shape_scores_formatted.size
        if axis_samples < 0:
            axis_samples += n_axes_scores_formatted

    n_samples = cc_stats_basics.scores_to_n_samples(scores_formatted, axis_samples, keepdims=keepdims)
    variances = cc_stats_basics.scores_to_variances(scores_formatted, axis_samples, keepdims=keepdims)

    shape_n_samples = np.asarray(n_samples.shape)
    n_axes_n_samples = shape_n_samples.size
    indexes_0 = np.empty(n_axes_n_samples, dtype=object)
    indexes_0[:] = slice(None)
    indexes_0[axis_comparisons] = 0

    indexes_1 = np.copy(indexes_0)
    indexes_1[axis_comparisons] = 1
    indexes_0_tuple = tuple(indexes_0)
    indexes_1_tuple = tuple(indexes_1)

    if equal_variances:
        # TO DO:
        # EQUATIONS ARE AT https://www.statsdirect.co.uk/help/parametric_methods/utt.htm
        denominator = None
    else:
        denominator = np.sqrt(
            (variances[indexes_0_tuple] / n_samples[indexes_0_tuple]) +
            (variances[indexes_1_tuple] / n_samples[indexes_1_tuple]))

    if keepdims:
        denominator = np.expand_dims(denominator, axis=axis_comparisons)

    return denominator


def scores_formatted_to_unpaired_t_test(
        scores_formatted, axis_comparisons, axis_samples, equal_variances=False, alpha=0.05, tails='2', keepdims=False):

    np.seterr(divide='ignore', invalid='ignore')

    confidence = 1 - alpha

    if scores_formatted.dtype == object:
        shape_object_scores_formatted = np.asarray(scores_formatted.shape)
        n_axes_object_scores_formatted = shape_object_scores_formatted.size
        # axes_object_scores_formatted = np.arange(n_axes_object_scores_formatted)
        if axis_comparisons < 0:
            axis_comparisons += n_axes_object_scores_formatted

        indexes_object = np.empty(n_axes_object_scores_formatted, dtype=object)
        indexes_object[:] = 0
        shape_scores_formatted = np.asarray(scores_formatted[tuple(indexes_object)].shape)
        n_axes_scores_formatted = shape_scores_formatted.size
        if axis_samples < 0:
            axis_samples += n_axes_scores_formatted
    else:
        shape_scores_formatted = np.asarray(scores_formatted.shape)
        n_axes_scores_formatted = shape_scores_formatted.size
        if axis_comparisons < 0:
            axis_comparisons += n_axes_scores_formatted
        if axis_samples < 0:
            axis_samples += n_axes_scores_formatted

    means_of_scores = cc_stats_basics.scores_to_means(scores_formatted, axis_samples, keepdims=keepdims)
    diff_of_means = cc_stats_basics.scores_to_diff_of_scores(means_of_scores, axis_comparisons, keepdims=keepdims)

    denominator = scores_formatted_to_denominator_of_unpaired_t_test(
        scores_formatted, axis_comparisons, axis_samples,
        equal_variances=equal_variances, keepdims=keepdims, formt=False)

    t_values = diff_of_means / denominator
    np.nan_to_num(t_values, copy=False, nan=0.0)

    df = scores_formatted_to_df_in_unpaired_t_test(
        scores_formatted, axis_comparisons, axis_samples,
        equal_variances=equal_variances, keepdims=keepdims, formt=False)

    shape_df = df.shape
    # n_df = df.size
    indexes_df = n_conditions_to_combinations(shape_df)

    t_critical = np.empty(shape_df, dtype=float)
    p_values = np.empty(shape_df, dtype=float)

    for indexes_df_i in indexes_df:
        indexes_df_i_tuple = tuple(indexes_df_i)
        if tails == '2':
            t_critical[indexes_df_i_tuple] = t.ppf((1 + confidence) / 2., df[indexes_df_i_tuple])
            p_values[indexes_df_i_tuple] = (1.0 - t.cdf(abs(t_values[indexes_df_i_tuple]), df[indexes_df_i_tuple])) * 2
        elif tails == '1l':
            t_critical[indexes_df_i_tuple] = -t.ppf(confidence, df[indexes_df_i_tuple])
            p_values[indexes_df_i_tuple] = t.cdf(t_values[indexes_df_i_tuple], df[indexes_df_i_tuple])
        elif tails == '1r':
            t_critical[indexes_df_i_tuple] = t.ppf(confidence, df[indexes_df_i_tuple])
            p_values[indexes_df_i_tuple] = 1.0 - t.cdf(t_values[indexes_df_i_tuple], df[indexes_df_i_tuple])

    return t_values, df, t_critical, p_values


# def height_of_t_confidence_interval(data, confidence=0.95):
#     a = 1.0 * np.array(data)
#     n = len(a)
#     m, se = np.mean(a), stats.sem(a)
#     h = se * stats.t.ppf((1 + confidence) / 2., n-1)
#     return h



from os.path import isfile as os_path_isfile
from time import sleep


def wait_downloading(directory_saved_as, max_seconds_wait=60):

    wait = True
    downloaded = False
    seconds_wait = 0
    while wait:
        if os_path_isfile(directory_saved_as):
            wait = False
            downloaded = True

        wait = wait and (seconds_wait < max_seconds_wait)
        sleep(1)
        seconds_wait += 1
            
    return downloaded



import os


def if_time_is_out(year, month, day, hour=0, minute=0, second=0, microsecond=0, restart=False):

    from datetime import datetime
    
    time_out = datetime(year, month, day, hour=hour, minute=minute, second=second, microsecond=microsecond)
    time_now = datetime.now()
    time_is_out = time_out <= time_now
    if time_is_out:

        if restart:
            os.system("shutdown /r /t 1")
        else:
            os.system("shutdown /s /t 1")

    return time_is_out


def now(restart=False):

    if restart:
        os.system("shutdown /r /t 1")
    else:
        os.system("shutdown /s /t 1")



from numbers import *
import numpy as np


def factors_of_x(x, y=1):
    type_x = type(x)
    type_y = type(y)

    x_is_integer = isinstance(x, Integral)

    # The Numeric abstract base classes:
    # numbers.Complex
    # numbers.Real
    # numbers.Integral
    # numbers.Number

    if not x_is_integer:
        raise Exception('x must be ab integer. Now, type_x = {}'.format(type_x))

    y_is_integer = isinstance(y, Integral)
    if not y_is_integer:
        raise Exception('y must be ab integer. Now, type_y = {}'.format(type_y))

    factors = np.empty(0, int)

    for i in range(y, x + 1):
        if x % i == 0:
            factors = np.append(factors, i)

    return factors


import numpy as np
from ccalafiore.combinations import n_conditions_to_combinations
from glob import glob
import numbers
from math import ceil


def advanced_indexing(indexes_loose):
    def clean_input_of_ix_(indexes_loose):
        n_axes = len(indexes_loose)
        for a in range(n_axes):
            try:
                len(indexes_loose[a])
            except TypeError:
                indexes_loose[a] = np.asarray([indexes_loose[a]], dtype=int)
        return indexes_loose

    indexes_loose = clean_input_of_ix_(indexes_loose)
    indexes = np.ix_(*indexes_loose)
    return indexes


def samples_in_arr1_are_in_arr2(arr1, arr2, axis=0):

    # OLD elements_in_x_are_in_y()

    arr1, arr2 = make_sure_is_iterable(arr1, arr2)

    arr1 = np.array(arr1)
    type_arr1 = arr1.dtype
    if type_arr1 == object:
        arr2 = np.array(list(arr1))

    arr2 = np.array(arr2)
    type_arr2 = arr2.dtype
    if type_arr2 == object:
        arr2 = np.array(list(arr2))

    shape_arr1 = np.array(arr1.shape)
    shape_arr2 = np.array(arr2.shape)

    n_axes_arr1 = len(shape_arr1)
    n_axes_arr2 = len(shape_arr2)
    if n_axes_arr1 != n_axes_arr2:

        if n_axes_arr1 == n_axes_arr2 - 1:

            arr1 = np.expand_dims(arr1, axis=axis)
            shape_arr1 = np.array(arr1.shape)
            n_axes_arr1 = len(shape_arr1)

        else:
            raise Exception('n_axes_arr1 has to be equal to n_axes_arr2\n'
                            'or to n_axes_arr2 - 1.\n'
                            'Now, n_axes_arr1 = {} and n_axes_arr2 = {}.'.format(
                n_axes_arr1, n_axes_arr2))

    axes_arr1_and_arr2 = np.arange(n_axes_arr1)

    if np.all(axis != axes_arr1_and_arr2):
        raise Exception('axis_1 is not in the axes_arr1_and_arr2.\n'
                        'axis_1 has to be equal to either one of the\n'
                        'axes_arr1_and_arr2.\n'
                        'Now, axis_1 = {} and axes_arr1_and_arr2 = {}.\n'
                        'By default, axis_1=0.'.format(
                         axis, axes_arr1_and_arr2))

    n_samples_arr1 = shape_arr1[axis]
    n_samples_arr2 = shape_arr2[axis]

    axes_within_samples = np.argwhere(axes_arr1_and_arr2 != axis)[:, 0]
    # dimensions_within_samples = dimensions_within_samples.astype(int)

    shape_1_sample_arr1 = shape_arr1[axes_within_samples]
    shape_1_sample_arr2 = shape_arr2[axes_within_samples]
    if np.any(shape_1_sample_arr1 != shape_1_sample_arr2):

        raise Exception('shape_1_sample_arr1 has to be equal to shape_1_sample_arr2.\n'
                        'Now,shape_1_sample_arr1 = {} and shape_1_samples_arr2 = {}.'.format(
            shape_1_sample_arr1, shape_1_sample_arr2))

    shape_arr_logical = np.copy(shape_arr2)
    shape_arr_logical = np.insert(shape_arr_logical, axis, shape_arr1[axis], axis=0)
    n_axes_arr_logical = len(shape_arr_logical)
    arr_logical = np.empty(shape_arr_logical, dtype=bool)

    indexes_arr1 = np.full(n_axes_arr1, slice(None))
    indexes_arr_logical = np.full(n_axes_arr_logical, slice(None))

    # indexes_arr1 = np.empty(n_dimensions_arr1, dtype=object)
    # indexes_arr_logical = np.empty(n_dimensions_arr_logical, dtype=object)
    # for i_dimension in range(n_dimensions_arr_logical):
    #     if i_dimension != axis_1:
    #         # indexes_samples_i_window.append(slice(0, n_samples_window))
    #     #     indexes_arr1[i_dimension] = np.arange([0])
    #     # else:
    #         indexes_arr_logical[i_dimension] = np.arange(shape_arr_logical[i_dimension])
    #         if i_dimension < n_dimensions_arr1:
    #             indexes_arr1[i_dimension] = np.arange(shape_arr1[i_dimension])

    for i_sample in range(n_samples_arr1):
        indexes_arr1[axis] = i_sample
        indexes_arr_logical[axis] = i_sample
        arr_logical[tuple(indexes_arr_logical)] = arr1[tuple(indexes_arr1)] == arr2

    elements_of_arr1_in_arr2_logical = np.any(arr_logical, axis=axis + 1)

    return elements_of_arr1_in_arr2_logical


def make_sure_is_iterable(*variables, if_not_iterable_convert_2_type_variable=np.ndarray):
    # list_types_iterable_allowed = [list, tuple, np.ndarray, str]

    # n_types_iterable_allowed = len(list_types_iterable_allowed)
    # if n_types_iterable_allowed == 1:
    #     if_not_iterable_convert_2_type_variable = list_types_iterable_allowed[0]
    # else:
    #     if if_not_iterable_convert_2_type_variable not in list_types_iterable_allowed:
    #         raise Exception(
    #             'if_not_iterable_convert_2_type_variable is not in list_types_iterable_allowed.\n'
    #             'Now, if_not_iterable_convert_2_type_variable = {} and list_types_iterable_allowed = {}.\n'
    #             'By default, make_sure_is_iterable(*variables, list_types_iterable_allowed=[list, tuple, np.ndarray],\n'
    #             'if_not_iterable_convert_2_type_variable=np.ndarray)'.format(
    #                 if_not_iterable_convert_2_type_variable, list_types_iterable_allowed))

    variables = list(variables)
    n_variables = len(variables)

    for i in range(n_variables):

        try:
            len(variables[i])

        except TypeError:

            if if_not_iterable_convert_2_type_variable == str:
                variables[i] = str(variables[i])

            else:
                variables[i] = [variables[i]]

                if if_not_iterable_convert_2_type_variable == list:
                    variables[i] = list(variables[i])

                elif if_not_iterable_convert_2_type_variable == tuple:
                    variables[i] = tuple(variables[i])

                elif if_not_iterable_convert_2_type_variable == np.ndarray:
                    variables[i] = np.array(variables[i])
                else:
                    raise Exception(
                        'if_not_iterable_convert_2_type_variable has to either be equal to np.ndarray, list, tuple or str.\n'
                        'Now, if_not_iterable_convert_2_type_variable = {}.\n'
                        'By default, if_not_iterable_convert_2_type_variable=np.ndarray'.format(
                            if_not_iterable_convert_2_type_variable))

        # type_i_variable = type(variables[i])
        #
        # # if type_i_variable not in list_types_iterable_allowed:
        #     if if_not_iterable_convert_2_type_variable == list:
        #         variables[i] = list([variables[i]])
        #
        #     elif if_not_iterable_convert_2_type_variable == tuple:
        #         variables[i] = tuple([variables[i]])
        #
        #     elif if_not_iterable_convert_2_type_variable == np.ndarray:
        #         variables[i] = np.array([variables[i]])
        #     else:
        #         raise Exception(
        #             'if_not_iterable_convert_2_type_variable has to either be equal to np.ndarray, list or tuple.\n'
        #             'Now, if_not_iterable_convert_2_type_variable = {}.\n'
        #             'By default, if_not_iterable_convert_2_type_variable=np.ndarray'.format(
        #                 if_not_iterable_convert_2_type_variable))

        # elif type_i_variable == np.ndarray:
        #     if variables[i].ndim == 0:
        #         variables[i] = np.array([variables[i]])

    if n_variables == 1:
        return variables[0]
    else:
        return variables


def transfer_n_random_samples_from_arr1_to_arr2(arr1, arr2, n_samples=1, axis=0, replace=False):
    # def transfer_n_random_samples_from_arr1_to_arr2(arr1, arr2, first_n_samples=1, all=False):

    # SPLIT THIS IN TWO FUNCTIONS:
    # 1) transfer_n_random_samples_from_arr1_to_arr2;
    # 2) transfer_n_samples_from_arr1_to_arr2;

    arr1, arr2 = make_sure_is_iterable(arr1, arr2)


    shape_arr1 = np.array(arr1.shape)
    shape_arr2 = np.array(arr2.shape)

    n_dimensions_arr1 = len(shape_arr1)
    n_dimensions_arr2 = len(shape_arr2)
    if n_dimensions_arr1 != n_dimensions_arr2:

        if n_dimensions_arr1 == n_dimensions_arr2 - 1:

            arr1 = np.expand_dims(arr1, axis=axis)
            shape_arr1 = np.array(arr1.shape)
            n_dimensions_arr1 = len(shape_arr1)

        else:
            raise Exception('n_dimensions_arr1 has to be equal to n_dimensions_arr2\n'
                            'or to n_dimensions_arr2 - 1.\n'
                            'Now, n_dimensions_arr1 = {} and n_dimensions_arr2 = {}.'.format(
                n_dimensions_arr1, n_dimensions_arr2))

    dimensions_arr1_and_arr2 = np.arange(n_dimensions_arr1)

    if np.all(axis != dimensions_arr1_and_arr2):
        raise Exception('axis is not in the dimensions_arr1_and_arr2.\n'
                        'axis has to be equal to either one ot the\n'
                        'dimensions_arr1_and_arr2.\n'
                        'Now, axis = {} and dimensions_arr1_and_arr2 = {}.\n'
                        'By default, axis=0.'.format(
            axis, dimensions_arr1_and_arr2))

    n_samples_arr1 = shape_arr1[axis]
    n_samples_arr2 = shape_arr2[axis]
    if n_samples > n_samples_arr1 and replace is False:
        raise Exception('n_samples cannot be greater than n_samples of arr1, when replace=False.\n'
                        'Now, n_samples = {}, n_samples_arr1 = {} and replace = {}.'.format(
            n_samples, n_samples_arr1, replace))


    dimensions_within_samples = np.argwhere(dimensions_arr1_and_arr2 != axis)[:, 0]
    # dimensions_within_samples = dimensions_within_samples.astype(int)

    shape_1_sample_arr1 = shape_arr1[dimensions_within_samples]
    shape_1_sample_arr2 = shape_arr2[dimensions_within_samples]
    if np.any(shape_1_sample_arr1 != shape_1_sample_arr2):

        raise Exception('shape_1_sample_arr1 has to be equal to shape_1_sample_arr2.\n'
                        'Now,shape_1_sample_arr1 = {} and shape_1_samples_arr2 = {}.'.format(
            shape_1_sample_arr1, shape_1_sample_arr2))

    indexes_chosen_samples_in_arr1 = np.random.choice(n_samples_arr1, n_samples, replace=replace)

    indexes = np.empty(n_dimensions_arr1, dtype=object)
    for i_d in range(n_dimensions_arr1):
        if i_d == axis:
            # indexes_samples_i_window.append(slice(0, n_samples_window))
            indexes[i_d] = indexes_chosen_samples_in_arr1
        else:
            indexes[i_d] = np.arange(shape_arr1[i_d])

    # if n_dimensions_arr1 == 1:
    #     chosen_sample_in_arr1 = arr1[np.ix_(indexes)]
    # else:
    chosen_sample_in_arr1 = arr1[np.ix_(*indexes)]

    arr2 = np.append(arr2, chosen_sample_in_arr1)

    arr1 = np.delete(arr1, indexes_chosen_samples_in_arr1, axis=axis)
    # indexes_chosen_sample_in_arr1 = np.argwhere(arr1 == chosen_sample_in_arr1)[0]

    # n_chosen_samples_in_arr1 = len(indexes_chosen_sample_in_arr1)
    # if all:
    #
    # elif n_chosen_samples_in_arr1 <= first_n_samples:
    #     indexes_chosen_sample_in_arr1 = indexes_chosen_sample_in_arr1[np.arange(n_chosen_samples_in_arr1)]
    #
    # elif n_chosen_samples_in_arr1 > first_n_samples:
    #     indexes_chosen_sample_in_arr1 = indexes_chosen_sample_in_arr1[np.arange(first_n_samples)]

    # arr1 = np.delete(arr1, indexes_chosen_sample_in_arr1)

    return arr1, arr2


def shuffle_in_windows(data, n_samples_window=None, n_windows=None, dimension_shuffle=0):

    type_data = type(data)
    if type_data != np.ndarray:
        raise Exception('data must be an numpy array. '
                        'Now, the class of data is {}'.format(type_data))


    dimension_shuffle_is_integer = isinstance(dimension_shuffle, numbers.Integral)
    # The Numeric abstract base classes:
    # numbers.Complex
    # numbers.Real
    # numbers.Integral
    # numbers.Number
    if not dimension_shuffle_is_integer:
        type_dimension_shuffle = type(dimension_shuffle)
        raise Exception('dimension_shuffle must be an integer. '
                        'Now, the class of dimension_shuffle is {}'.format(
            type_dimension_shuffle))

    shape_data = data.shape
    n_samples_data = shape_data[dimension_shuffle]
    n_d_data = len(shape_data)
    if dimension_shuffle > n_d_data:
        raise Exception('shuffle in dimension {} can\'t be done,\n'
                        'because data has {} dimensions and\n'
                        'it doesn\'t have the {}th dimension.\n'
                        'dimension_shuffle has to be an integer in the interval\n'
                        '0 =< dimension_shuffle < n_d_data.\n'
                        'In this case, dimension_shuffle = {} and n_d_data = {}'.format(
            dimension_shuffle, n_d_data, dimension_shuffle, dimension_shuffle, n_d_data))


    n_samples_window_is_integer = isinstance(n_samples_window, numbers.Integral)

    if n_windows is None and not n_samples_window_is_integer:
        type_n_samples_window = type(n_samples_window)
        raise Exception('n_samples_window must be an integer. '
                        'Now, the class of n_samples_window is {}'.format(
            type_n_samples_window))

    n_windows_is_integer = isinstance(n_windows, numbers.Integral)
    if n_samples_window is None and not n_windows_is_integer:
        type_n_windows = type(n_windows)
        raise Exception('n_windows must be an integer. '
                        'Now, the class of n_windows is {}'.format(
            type_n_windows))

    if n_samples_window is None and n_windows is None:
        raise Exception('Both n_samples_window and n_windows are None.'
                        'Define only one of them with an integer.'
                        'One of them has to be None and the other one has to be an integer')

    if n_samples_window is None and n_windows_is_integer:
        n_samples_window = ceil(n_samples_data / n_windows)

    if n_windows is None and n_samples_window_is_integer:
        n_windows = ceil(n_samples_data / n_samples_window)

    if n_samples_window_is_integer and n_windows_is_integer:

        tmp_n_samples_window = ceil(n_samples_data / n_windows)
        tmp_n_windows = ceil(n_samples_data / n_samples_window)

        if tmp_n_samples_window != n_samples_window and tmp_n_windows != n_windows:

            raise Exception('For n_samples_data = {}, n_samples_window = {} and n_windows = {} are not compatible.\n'
                            'If n_samples_window = {}, n_windows has to be {} or None.\n'
                            'If n_windows = {}, n_samples_window has to be {} or None'.format(
                n_samples_data, n_samples_window, n_windows, n_samples_window, tmp_n_windows, n_windows, tmp_n_samples_window))




    if n_samples_window * n_windows > n_samples_data:
        print('Warning: n_samples_window * n_windows > n_samples_data.\n'
              'n_samples_window * n_windows = {} and n_samples_data = {}'.format(
            n_samples_window * n_windows, n_samples_data))

    indexes_samples_i_window = np.empty(n_d_data, dtype=object)
    for i_d in range(n_d_data):
        if i_d == dimension_shuffle:
            # indexes_samples_i_window.append(slice(0, n_samples_window))
            indexes_samples_i_window[i_d] = np.arange(n_samples_window)
        else:
            indexes_samples_i_window[i_d] = np.arange(shape_data[i_d])

    # indexes_samples_i_window = np.array(indexes_samples_i_window)
    # indexes_samples_i_window[dimension_shuffle] = indexes_samples_i_window[dimension_shuffle][indexes_samples_i_window[dimension_shuffle] < n_samples_data]
    # indexes_samples_i_window_tuple = np.ix_(*indexes_samples_i_window)

    tmp_shape_data_shuffle = list(shape_data)
    tmp_shape_data_shuffle[dimension_shuffle] = 0
    dtype_data = data.dtype
    # data_shuffle = np.empty(tmp_shape_data_shuffle, dtype=dtype_data)
    data_shuffle = np.empty(shape_data, dtype=dtype_data)

    for i_window in range(n_windows):

        indexes_samples_i_window[dimension_shuffle] = indexes_samples_i_window[dimension_shuffle][indexes_samples_i_window[dimension_shuffle] < n_samples_data]
        # indexes_samples_i_window = np.arange(n_samples_window * i_window, n_samples_window * (i_window + 1))

        # indexes_samples_i_window = indexes_samples_i_window[indexes_samples_i_window < n_samples_data]
        # index_tuple = tuple(index)
        # index[dimension_shuffle] = start_i_window
        # data = np.arange(6*6*6).reshape(6, 6, 6)
        # i = np.ix_(*[[1, 3], [1, 3, 5], [0, 1, 3, 5]])
        # if n_d_data == 1:
        #     samples_i_window = data[np.ix_(indexes_samples_i_window)]
        # else:
        indexes_samples_i_window_tuple = np.ix_(*indexes_samples_i_window)
            # samples_i_window = data[np.ix_(indexes_samples_i_window_tuple)]
        samples_i_window = data[indexes_samples_i_window_tuple]
            # samples_i_window = np.take(data, indexes_samples_i_window, axis=dimension_shuffle)

        samples_i_window = shuffle_in_any_dimension(samples_i_window, dimension_shuffle=dimension_shuffle)
        # data_shuffle = np.append(data_shuffle, samples_i_window, axis=dimension_shuffle)
        data_shuffle[indexes_samples_i_window_tuple] = samples_i_window

        # indexes_samples_i_window = np.array(indexes_samples_i_window_tuple)
        # indexes_samples_i_window[dimension_shuffle] += n_samples_window
        # indexes_samples_i_window[dimension_shuffle] = indexes_samples_i_window[dimension_shuffle][indexes_samples_i_window[dimension_shuffle] < n_samples_data]
        # indexes_samples_i_window_tuple = tuple(indexes_samples_i_window)

        if i_window < n_windows - 1:
            indexes_samples_i_window[dimension_shuffle] += n_samples_window


    return data_shuffle


def shuffle_in_any_dimension(data, axis=0):

    type_data = type(data)
    if type_data != np.ndarray:
        raise Exception('data must be an numpy array. '
                        'Now, the class of data is {}'.format(type_data))

    dimension_shuffle_is_integer = isinstance(axis, numbers.Integral)
    if not dimension_shuffle_is_integer:
        type_dimension_shuffle = type(axis)
        raise Exception('dimension_shuffle must be an integer. '
                        'Now, the class of dimension_shuffle is {}'.format(
            type_dimension_shuffle))

    shape_data = data.shape
    n_samples = shape_data[axis]
    n_axes = len(shape_data)

    if axis > n_axes:
        raise Exception('shuffle in axis {} can\'t be done,\n'
                        'because data has {} axes and\n'
                        'it doesn\'t have the {}th axis.\n'
                        'axis has to be an integer in the interval\n'
                        '0 =< axis < n_axes.\n'
                        'In this case, axis = {} and n_axes = {}'.format(
            axis, n_axes, axis, axis, n_axes))


    indexes_new = np.empty(n_axes, dtype=object)
    for a in range(n_axes):
        if a == axis:
            # indexes_samples_i_window.append(slice(0, n_samples_window))
            indexes_new[a] = np.random.rand(n_samples).argsort()
        else:
            indexes_new[a] = np.arange(shape_data[a])

    # indexes_new = []
    # for i_d in range(n_d_data):
    #     if i_d == dimension_shuffle:
    #         # indexes_samples_i_window.append(slice(0, n_samples_window))
    #         indexes_new.append(np.random.rand(n_samples).argsort())
    #     else:
    #         indexes_new.append(np.arange(0, shape_data[i_d]))
    #
    # indexes_new = np.array(indexes_new)

    # indexes_new = np.random.rand(n_samples).argsort()
    # data_shuffle = np.take(data, indexes_new, axis=dimension_shuffle, out=data)
    # if n_d_data == 1:
    #     data_shuffle = data[np.ix_(indexes_new)]
    #
    # else:
    data_shuffle = data[advanced_indexing(indexes_new)]

    return data_shuffle


def n_csv_files_to_array(directories_csv_files, names_csv_files_in_directories=False, rows=slice(None), columns=slice(None), data_type=None):

    array = None

    n_axes_directories = len(directories_csv_files)
    n_conditions_directories = np.empty(n_axes_directories, dtype=int)
    for a in range(n_axes_directories):
        n_conditions_directories[a] = len(directories_csv_files[a])

    logical_indexes_conditions = n_conditions_directories > 1

    combinations_directories = n_conditions_to_combinations(n_conditions_directories)
    n_combinations_directories = combinations_directories.shape[0]

    for d in range(n_combinations_directories):

        directory = directories_csv_files[0][combinations_directories[d, 0]]
        for a in range(1, n_axes_directories):

            directory = '/'.join([
                directory,
                directories_csv_files[a][combinations_directories[d, a]]
            ])

        if names_csv_files_in_directories:
            files_per_directory = [directory]
        else:
            files_per_directory = glob('/'.join([directory, '*.csv']))

        n_files_per_directory = len(files_per_directory)

        start_row_f_file = 0
        for f in range(n_files_per_directory):

            array_per_file = np.genfromtxt(files_per_directory[f], delimiter=',', dtype=data_type)[rows, columns]

            if array is None:

                shape_array_per_file = np.asarray(array_per_file.shape, dtype=int)

                shape_array = np.asarray([
                    *n_conditions_directories[logical_indexes_conditions],
                    n_files_per_directory * shape_array_per_file[0],
                    *shape_array_per_file[1:]
                ], dtype=int)

                array = np.empty(shape_array, dtype=data_type)

            end_row_f_file = (f + 1) * shape_array_per_file[0]

            indexes_array = tuple([
                *combinations_directories[d][logical_indexes_conditions],
                slice(start_row_f_file, end_row_f_file)
            ])

            array[indexes_array] = array_per_file
            start_row_f_file = end_row_f_file

    return array


def split(array, n_slices, axis=0):

    shape_array = np.array(array.shape)

    n_dimensions_array = len(shape_array)

    indexes = np.empty(n_dimensions_array, dtype=object)

    for i_d in range(n_dimensions_array):
        if i_d != axis:

            indexes[i_d] = slice(None)

    # indexes_samples_i_window.append(slice(0, n_samples_window))

    n_elements_array = shape_array[axis]

    n_elements_slice = n_elements_array / n_slices

    sliced_array = np.empty(n_slices, dtype=object)

    index_start = 0
    for i_slice in range(n_slices):

        index_stop = int((i_slice + 1) * n_elements_slice)

        indexes[axis] = slice(index_start, index_stop)

        # sliced_array[i_slice] = array[int(i_slice * n_elements_slice): int((i_slice + 1) * n_elements_slice)]
        sliced_array[i_slice] = array[tuple(indexes)]

        index_start = index_stop

    return sliced_array


def pad_array_from_n_samples_target(array, n_samples_target=1, axis=0):

    try:
        len(array)
        array = np.asarray(array)
    except TypeError:
        array = np.asarray([array])

    shape_array = np.asarray(array.shape)
    n_axes = len(shape_array)
    while n_axes <= axis:
        array = np.expand_dims(array, axis=n_axes)
        shape_array = np.asarray(array.shape)
        n_axes = len(shape_array)

    n_samples_true = shape_array[axis]
    while n_samples_true < n_samples_target:
        array = np.append(array, array, axis=axis)
        n_samples_true = array.shape[axis]

    index_array = np.full(n_axes, slice(None))
    index_array[axis] = slice(n_samples_target)
    array = array[tuple(index_array)]

    return array


def pad_arr_1_from_arr_2(arr_1, arr_2, axis=0):

    try:
        len(arr_1)
        arr_1 = np.asarray(arr_1)
    except TypeError:
        arr_1 = np.asarray([arr_1])

    try:
        len(arr_2)
        arr_2 = np.asarray(arr_2)
    except TypeError:
        arr_2 = np.asarray([arr_2])

    shape_arr_1 = np.asarray(arr_1.shape)
    n_axes_1 = len(shape_arr_1)
    shape_arr_2 = np.asarray(arr_2.shape)
    n_axes_2 = len(shape_arr_2)

    if n_axes_1 == n_axes_2 - 1:
        arr_1 = np.expand_dims(arr_1, axis=axis)
        shape_arr_1 = np.asarray(arr_1.shape)
        n_axes_1 = len(shape_arr_1)
    elif n_axes_1 == n_axes_2:
        pass
    else:
        raise Exception('dimensions of arr_1 and arr_2 do not match')

    if np.all(shape_arr_1[np.arange(n_axes_1) != axis] == shape_arr_2[np.arange(n_axes_2) != axis]):
        pass
    else:
        raise Exception('dimensions of arr_1 and arr_2 do not match')

    n_samples_true = shape_arr_1[axis]
    n_samples_target = shape_arr_2[axis]

    while n_samples_true < n_samples_target:
        arr_1 = np.append(arr_1, arr_1, axis=axis)
        n_samples_true = arr_1.shape[axis]

    index_arr_1 = np.full(n_axes_1, slice(None))
    index_arr_1[axis] = slice(n_samples_target)

    arr_1 = arr_1[tuple(index_arr_1)]

    return arr_1



import numpy as np
from ccalafiore.array import pad_arr_1_from_arr_2, samples_in_arr1_are_in_arr2, advanced_indexing
from scipy.stats import t, sem
from scipy.stats.stats import _ttest_finish
from ccalafiore.combinations import n_conditions_to_combinations


def scores_to_df_for_unpaired_t_test(
        scores, axis_samples=-1, axis_independent_variables=-2, keepdims=False):

    shape_scores = np.asarray(scores.shape)
    n_axes_scores = len(shape_scores)
    axis_samples %= n_axes_scores
    axis_independent_variables_scores = axis_independent_variables % n_axes_scores

    n_samples = np.sum(np.logical_not(np.isnan(scores)), axis=axis_samples, keepdims=keepdims)
    variances = scores_to_variances(scores, axis_samples=axis_samples, keepdims=keepdims)
    if not(keepdims):
        axis_independent_variables_n_samples = axis_independent_variables_scores - (
                axis_independent_variables_scores > axis_samples)
    else:
        axis_independent_variables_n_samples = axis_independent_variables_scores

    n_samples = np.expand_dims(n_samples, axis=axis_independent_variables_n_samples + 1)
    variances = np.expand_dims(variances, axis=axis_independent_variables_n_samples + 1)
    shape_n_samples = np.asarray(n_samples.shape)
    n_axes_n_samples = len(shape_n_samples)
    n_conditions_iv = shape_n_samples[axis_independent_variables_n_samples]

    axes_n_samples = np.arange(n_axes_n_samples)
    axes_non_independent_variables_n_samples = axes_n_samples[np.logical_not(samples_in_arr1_are_in_arr2(
        axes_n_samples, axis_independent_variables_n_samples))]

    index_n_samples_1 = np.empty(n_axes_n_samples, dtype=object)
    index_df_1 = np.copy(index_n_samples_1)

    for a in axes_non_independent_variables_n_samples:
        index_n_samples_1[a] = np.arange(shape_n_samples[a])
    index_n_samples_2 = np.copy(index_n_samples_1)

    axis_independent_variables_df = np.asarray(
        [axis_independent_variables_n_samples, axis_independent_variables_n_samples + 1], dtype=int)
    axes_non_independent_variables_df = axes_n_samples[np.logical_not(samples_in_arr1_are_in_arr2(
        axes_n_samples, axis_independent_variables_df))]

    index_df_1[axes_non_independent_variables_df] = index_n_samples_1[axes_non_independent_variables_df]
    index_df_2 = np.copy(index_df_1)

    shape_df = np.copy(shape_n_samples)
    shape_df[axis_independent_variables_df] = n_conditions_iv

    df = np.full(shape_df, 0, dtype=int)

    for i_condition_1 in range(n_conditions_iv):

        index_n_samples_1[axis_independent_variables_n_samples] = i_condition_1
        index_n_samples_1_adv = advanced_indexing(index_n_samples_1)
        index_df_1[axis_independent_variables_df[0]] = i_condition_1
        index_df_2[axis_independent_variables_df[1]] = i_condition_1

        for i_condition_2 in range(i_condition_1, n_conditions_iv):

            index_n_samples_2[axis_independent_variables_n_samples] = i_condition_2
            index_n_samples_2_adv = advanced_indexing(index_n_samples_2)

            index_df_1[axis_independent_variables_df[1]] = i_condition_2
            index_df_1_adv = advanced_indexing(index_df_1)

            variances_1 = variances[index_n_samples_1_adv]
            variances_2 = variances[index_n_samples_2_adv]

            n_samples_1 = n_samples[index_n_samples_1_adv]
            n_samples_2 = n_samples[index_n_samples_2_adv]

            df_1_2 = (((variances_1 / n_samples_1) + (variances_2 / n_samples_2))**2) / (
                (((variances_1 / n_samples_1)**2) / (n_samples_1 - 1)) +
                (((variances_2 / n_samples_2)**2) / (n_samples_2 - 1)))
            # df_1_2 = np.maximum(
            #     n_samples[index_n_samples_1_adv], n_samples[index_n_samples_2_adv])

            df[index_df_1_adv] = df_1_2

            if i_condition_1 < i_condition_2:
                index_df_2[axis_independent_variables_df[0]] = i_condition_2
                index_df_2_adv = advanced_indexing(index_df_2)
                df[index_df_2_adv] = df_1_2

    return df


def scores_to_unpaired_t_test_denominator(
        scores, axis_samples=-1, axis_independent_variables=-2, keepdims=False, dtype=None):

    shape_scores = np.asarray(scores.shape)
    n_axes_scores = len(shape_scores)
    axis_samples %= n_axes_scores
    axis_independent_variables %= n_axes_scores

    n_samples = np.sum(np.logical_not(np.isnan(scores)), axis=axis_samples, keepdims=keepdims)
    variances = scores_to_variances(scores, axis_samples=axis_samples, keepdims=keepdims)

    if not(keepdims):
        axis_independent_variables -= (axis_independent_variables > axis_samples)
    n_samples = np.expand_dims(n_samples, axis=axis_independent_variables + 1)
    variances = np.expand_dims(variances, axis=axis_independent_variables + 1)
    shape_variances = np.asarray(variances.shape)
    n_axes_variances = len(shape_variances)
    n_conditions_iv = shape_variances[axis_independent_variables]

    axes_variances = np.arange(n_axes_variances)
    axes_non_independent_variables_variances = axes_variances[np.logical_not(samples_in_arr1_are_in_arr2(
        axes_variances, axis_independent_variables))]

    index_1 = np.empty(n_axes_variances, dtype=object)
    index_denominator_1 = np.copy(index_1)

    for a in axes_non_independent_variables_variances:
        index_1[a] = np.arange(shape_variances[a])
    index_2 = np.copy(index_1)

    independent_variables_denominator = np.asarray([axis_independent_variables, axis_independent_variables + 1], dtype=int)
    axes_non_independent_variables_denominator = axes_variances[np.logical_not(samples_in_arr1_are_in_arr2(
        axes_variances, independent_variables_denominator))]

    index_denominator_1[axes_non_independent_variables_denominator] = index_1[axes_non_independent_variables_denominator]
    index_denominator_2 = np.copy(index_denominator_1)

    shape_denominator = np.copy(shape_variances)
    shape_denominator[independent_variables_denominator] = n_conditions_iv
    if dtype is None:
        dtype = variances.dtype
    denominator = np.full(shape_denominator, 0, dtype=dtype)

    for i_condition_1 in range(n_conditions_iv - 1):

        index_1[axis_independent_variables] = i_condition_1
        index_1_adv = advanced_indexing(index_1)
        index_denominator_1[independent_variables_denominator[0]] = i_condition_1
        index_denominator_2[independent_variables_denominator[1]] = i_condition_1

        for i_condition_2 in range(i_condition_1 + 1, n_conditions_iv):

            index_2[axis_independent_variables] = i_condition_2
            index_2_adv = advanced_indexing(index_2)

            index_denominator_1[independent_variables_denominator[1]] = i_condition_2
            index_denominator_1_adv = advanced_indexing(index_denominator_1)

            index_denominator_2[independent_variables_denominator[0]] = i_condition_2
            index_denominator_2_adv = advanced_indexing(index_denominator_2)

            denominator_1_2 = np.sqrt(
                (variances[index_1_adv] / n_samples[index_1_adv]) +
                (variances[index_2_adv] / n_samples[index_2_adv])
            )

            denominator[index_denominator_1_adv] = denominator_1_2
            denominator[index_denominator_2_adv] = denominator_1_2

    return denominator


def scores_to_unpaired_t_test(
        scores, axis_samples=-1, axis_independent_variables=-2, alpha=0.05, tails='2', keepdims=False):

    np.seterr(divide='ignore', invalid='ignore')

    confidence = 1 - alpha

    shape_scores = scores.shape
    n_axes = len(shape_scores)
    axis_samples %= n_axes
    axis_independent_variables %= n_axes

    means_of_scores = scores_to_means(scores, axis_samples=axis_samples, keepdims=keepdims)

    denominator = scores_to_unpaired_t_test_denominator(
        scores, axis_samples=axis_samples, axis_independent_variables=axis_independent_variables)

    df = scores_to_df_for_unpaired_t_test(
        scores, axis_samples=axis_samples, axis_independent_variables=axis_independent_variables,
        keepdims=keepdims)

    if not(keepdims):
        axis_independent_variables -= (axis_independent_variables > axis_samples)

    diff_of_means = scores_to_diff_of_scores(means_of_scores, axis_independent_variables=axis_independent_variables)

    t_values = diff_of_means / denominator
    np.nan_to_num(t_values, copy=False, nan=0.0)

    shape_df = df.shape
    n_df = df.size

    t_critical = np.empty(shape_df, dtype=float)
    p_values = np.empty(shape_df, dtype=float)

    indexs_df = n_conditions_to_combinations(shape_df)
    for i in range(n_df):

        if tails == '2':
            t_critical[tuple(indexs_df[i])] = t.ppf((1 + confidence) / 2., df[tuple(indexs_df[i])])
            p_values[tuple(indexs_df[i])] = (1.0 - t.cdf(abs(t_values[tuple(indexs_df[i])]), df[tuple(indexs_df[i])])) * 2
        elif tails == '1l':
            t_critical[tuple(indexs_df[i])] = -t.ppf(confidence, df[tuple(indexs_df[i])])
            p_values[tuple(indexs_df[i])] = t.cdf(t_values[tuple(indexs_df[i])], df[tuple(indexs_df[i])])
        elif tails == '1r':
            t_critical[tuple(indexs_df[i])] = t.ppf(confidence, df[tuple(indexs_df[i])])
            p_values[tuple(indexs_df[i])] = 1.0 - t.cdf(t_values[tuple(indexs_df[i])], df[tuple(indexs_df[i])])

    return t_values, df, t_critical, p_values


# def height_of_t_confidence_interval(data, confidence=0.95):
#     a = 1.0 * np.array(data)
#     n = len(a)
#     m, se = np.mean(a), stats.sem(a)
#     h = se * stats.t.ppf((1 + confidence) / 2., n-1)
#     return h



import numpy as np
from numbers import *


def n_conditions_to_combinations(
        n_conditions,
        axis_combinations=0,
        n_repetitions_combinations=1,
        order_variables='rl',
        variables_in_order=None,
        dtype=int):

    # order_variables is the order of the variables by which they change their own conditions.
    # order_variables can either be 'rl' for right-to-left, 'lr' for left-to-right or 'c' for custom.
    # If order_variables=='c', then variables_in_order must be a list, a tuple or an 1-d array containing the variables
    # in the custom order by which the variables change their own conditions:
    #
    # Requirements:
    # 1) order_variables == 'rl' or order_variables == 'lr' or order_variables == 'c';
    # If order_variables=='c', then:
    #     2) len(variables_in_order) == len(n_conditions);
    #     3) variables_in_order must contain all integers from 0 to (len(n_conditions)-1);
    #     4) variables_in_order cannot contain repeated integers.

    n_variables = len(n_conditions)
    if n_variables > 0:
        from ccalafiore.array import pad_array_from_n_samples_target
        variables = np.arange(n_variables)
        # n_conditions = np.asarray(n_conditions, dtype=int)
        # conditions = n_conditions_to_conditions(n_conditions)
        n_combinations = np.prod(n_conditions) * n_repetitions_combinations
        if n_combinations < 0:
            n_conditions = n_conditions.astype(np.int64)
            n_combinations = np.prod(n_conditions) * n_repetitions_combinations

        # axis_combinations = int(not(bool(axis_variables)))
        axis_variables = int(not(bool(axis_combinations)))
        shape_combinations = np.empty(2, dtype=int)
        shape_combinations[axis_combinations] = n_combinations
        shape_combinations[axis_variables] = n_variables

        combinations = np.empty(shape_combinations, dtype=dtype)

        if order_variables == 'rl':
            variables_in_order = variables[::-1]
        elif order_variables == 'lr':
            variables_in_order = variables
        elif order_variables == 'c':
            variables_in_order = np.asarray(variables_in_order, dtype=int)
            # check requirement 2
            if len(variables_in_order) != n_variables:
                raise ValueError('The following condition must be met:\nlen(variables_in_order) == len(n_conditions).')
            # check requirement 3
            for v in range(n_variables):
                if v not in variables_in_order:
                    raise ValueError('variables_in_order must contain all integers from 0 to (len(n_conditions)-1).\n'
                                     '{} is not in variables_in_order.'.format(v))
            # check requirement 4
            for v in variables_in_order:
                if np.sum(v == variables_in_order) > 1:
                    raise ValueError('variables_in_order cannot contain repeated integers.\n'
                                     '{} is a repeated integer.'.format(v))
        else:
            # check requirement 1
            raise ValueError(
                'order_variables can either be \'rl\' for right-to-left, '
                '\'lr\' for left-to-right or \'c\' for custom.\n'
                'Now, order_variables = {}'.format(order_variables))

        i_variable = variables_in_order[0]
        indexes_combinations = np.empty(2, dtype=object)
        indexes_combinations[axis_variables] = i_variable
        for i_condition in range(n_conditions[i_variable]):
            indexes_combinations[axis_combinations] = slice(i_condition, n_combinations, n_conditions[i_variable])
            combinations[tuple(indexes_combinations)] = i_condition

        indexes_combinations[axis_combinations] = slice(None)

        # combinations[tuple(indexes_combinations)] = np.expand_dims(
        #     pad_array_from_n_samples_target(conditions[-1], n_samples_target=n_combinations),
        #     axis=axis_variables)
        cumulative_n_combinations = n_conditions[i_variable]
        for i_variable in variables_in_order[1:]:
            cumulative_combinations = np.empty(
                cumulative_n_combinations * n_conditions[i_variable], combinations.dtype)
            for i_condition in range(n_conditions[i_variable]):
                cumulative_combinations[
                    slice(i_condition * cumulative_n_combinations,
                          (i_condition + 1) * cumulative_n_combinations)] = i_condition

            indexes_combinations[axis_variables] = i_variable
            if cumulative_combinations.size < n_combinations:
                combinations[tuple(indexes_combinations)] = pad_array_from_n_samples_target(
                    cumulative_combinations, n_samples_target=n_combinations)
            elif cumulative_combinations.size == n_combinations:
                combinations[tuple(indexes_combinations)] = cumulative_combinations
            else:
                raise Exception('something wrong')
            cumulative_n_combinations *= n_conditions[i_variable]

        # if change_dtype:
        #     combinations = combinations.astype(dtype_end)
    else:
        combinations = []
    return combinations


def conditions_to_combinations(
        conditions,
        axis_combinations=0,
        n_repetitions_combinations=1,
        order_variables='rl',
        variables_in_order=None,
        dtype=None):

    # order_variables is the order of the variables by which they change their own conditions.
    # order_variables can either be 'rl' for right-to-left, 'lr' for left-to-right or 'c' for custom.
    # If order_variables=='c', then variables_in_order must be a list, a tuple or an 1-d array containing the variables
    # in the custom order by which the variables change their own conditions:
    #
    # Requirements:
    # 1) order_variables == 'rl' or order_variables == 'lr' or order_variables == 'c';
    # If order_variables=='c', then:
    #     2) len(variables_in_order) == len(conditions);
    #     3) variables_in_order contains all integers from 0 to (len(conditions)-1);
    #     4) variables_in_order cannot contain repeated integers.

    n_variables = len(conditions)
    if n_variables > 0:
        from ccalafiore.array import pad_array_from_n_samples_target
        variables = np.arange(n_variables)
        # conditions = np.asarray(conditions, dtype=object)
        n_conditions = conditions_to_n_conditions(conditions)
        n_combinations = np.prod(n_conditions) * n_repetitions_combinations
        if n_combinations < 0:
            n_conditions = n_conditions.astype(np.int64)
            n_combinations = np.prod(n_conditions) * n_repetitions_combinations

        if dtype is None:
            dtype = type(conditions[0][0])

        change_dtype = False
        if dtype == str:
            dtype_end = str
            dtype = object
            change_dtype = True

        axis_variables = int(not(bool(axis_combinations)))
        shape_combinations = np.empty(2, dtype=int)
        shape_combinations[axis_combinations] = n_combinations
        shape_combinations[axis_variables] = n_variables

        combinations = np.empty(shape_combinations, dtype=dtype)

        if order_variables == 'rl':
            variables_in_order = variables[::-1]
        elif order_variables == 'lr':
            variables_in_order = variables
        elif order_variables == 'c':
            variables_in_order = np.asarray(variables_in_order, dtype=int)
            # check requirement 2
            if len(variables_in_order) != n_variables:
                raise ValueError('The following condition must be met:\nlen(variables_in_order) == len(conditions).')
            # check requirement 3
            for v in range(n_variables):
                if v not in variables_in_order:
                    raise ValueError('variables_in_order must contain all integers from 0 to (len(conditions)-1).\n'
                                     '{} is not in variables_in_order.'.format(v))
            # check requirement 4
            for v in variables_in_order:
                if np.sum(v == variables_in_order) > 1:
                    raise ValueError('variables_in_order cannot contain repeated integers.\n'
                                     '{} is a repeated integer.'.format(v))
        else:
            # check requirement 1
            raise ValueError(
                'order_variables can either be \'rl\' for right-to-left, '
                '\'lr\' for left-to-right or \'c\' for custom.\n'
                'Now, order_variables = {}'.format(order_variables))

        i_variable = variables_in_order[0]
        indexes_combinations = np.empty(2, dtype=object)
        indexes_combinations[axis_variables] = i_variable
        for i_condition in range(n_conditions[i_variable]):
            indexes_combinations[axis_combinations] = slice(i_condition, n_combinations, n_conditions[i_variable])
            combinations[tuple(indexes_combinations)] = conditions[i_variable][i_condition]

        indexes_combinations[axis_combinations] = slice(None)

        # combinations[tuple(indexes_combinations)] = np.expand_dims(
        #     pad_array_from_n_samples_target(conditions[-1], n_samples_target=n_combinations),
        #     axis=axis_variables)
        cumulative_n_combinations = n_conditions[i_variable]
        for i_variable in variables_in_order[1:]:
            cumulative_combinations = np.empty(
                cumulative_n_combinations * n_conditions[i_variable], combinations.dtype)
            for i_condition in range(n_conditions[i_variable]):
                cumulative_combinations[
                    slice(i_condition * cumulative_n_combinations,
                          (i_condition + 1) * cumulative_n_combinations)] = conditions[i_variable][i_condition]

            indexes_combinations[axis_variables] = i_variable
            if cumulative_combinations.size < n_combinations:
                combinations[tuple(indexes_combinations)] = pad_array_from_n_samples_target(
                    cumulative_combinations, n_samples_target=n_combinations)
            elif cumulative_combinations.size == n_combinations:
                combinations[tuple(indexes_combinations)] = cumulative_combinations
            else:
                raise Exception('something wrong')
            cumulative_n_combinations *= n_conditions[i_variable]

        if change_dtype:
            combinations = combinations.astype(dtype_end)
    else:
        combinations = []
    return combinations


def make_combinations_of_conditions_as_distributions(conditions_as_distributions, n_repetitions_combinations=1):

    n_variables = len(conditions_as_distributions)
    n_conditions = np.empty(n_variables, dtype=object)

    for i_variable in range(n_variables):

        n_conditions[i_variable] = len(conditions_as_distributions[i_variable])

    combinations_of_conditions = n_conditions_to_combinations(n_conditions)

    n_combinations = len(combinations_of_conditions)

    combinations_of_conditions_as_distributions = np.empty([n_combinations, n_variables], dtype=int)

    for i_variable in range(n_variables):

        for i_condition in range(n_conditions[i_variable]):

            indexes_of_i_condition = np.argwhere(combinations_of_conditions[:, i_variable] == i_condition)

            n_i_condition = len(indexes_of_i_condition)

            combinations_of_conditions_as_distributions[indexes_of_i_condition, i_variable] = \
                np.random.choice(conditions_as_distributions[i_variable][i_condition],
                                 n_i_condition)[:, None]

    return combinations_of_conditions_as_distributions


def trials_to_combinations(
        trials, axis_combinations=0, variables=None,  n_repetitions_combinations=1, dtype=None):

    conditions = trials_to_conditions(trials, axis_combinations=axis_combinations, variables=variables)
    combinations = conditions_to_combinations(
        conditions,
        axis_combinations=axis_combinations,
        n_repetitions_combinations=n_repetitions_combinations,
        dtype=dtype)
    return combinations


def trials_to_conditions(trials, axis_combinations=0, variables=None):

    from ccalafiore.array import advanced_indexing

    axis_variables = int(not(bool(axis_combinations)))
    if variables is None:
        n_variables = trials.shape[axis_variables]
        variables = np.arange(n_variables)
    else:
        try:
            n_variables = len(variables)
        except TypeError:
            variables = np.asarray([variables], dtype=int)
            n_variables = len(variables)

    conditions = np.empty(n_variables, dtype=object)

    indexes_trials_adv = np.empty(2, dtype=object)
    indexes_trials_adv[axis_combinations] = np.arange(trials.shape[axis_combinations])

    indexes_trials_slice = np.full(2, 0, dtype=object)
    indexes_trials_slice[axis_combinations] = slice(None)
    indexes_trials_slice = tuple(indexes_trials_slice)

    for i_variable in range(n_variables):

        indexes_trials_adv[axis_variables] = variables[i_variable]
        trials_variables_i = trials[advanced_indexing(indexes_trials_adv)]
        trials_variables_i = trials_variables_i[indexes_trials_slice]

        conditions[i_variable] = np.unique(trials_variables_i)

    return conditions


def trials_to_n_conditions(trials, axis_combinations=0, variables=None):
    conditions = trials_to_conditions(trials, axis_combinations=axis_combinations, variables=variables)
    n_conditions = conditions_to_n_conditions(conditions)
    # n_variables = len(conditions)
    # n_conditions = np.empty(n_variables, dtype=int)
    # for i_variable in range(n_variables):
    #     n_conditions[i_variable] = len(conditions[i_variable])
    return n_conditions


def n_conditions_to_conditions(n_conditions):
    n_variables = len(n_conditions)
    conditions = np.empty(n_variables, dtype=object)
    for i_variable in range(n_variables):
        conditions[i_variable] = np.arange(n_conditions[i_variable])
    return conditions


def conditions_to_n_conditions(conditions):
    n_variables = len(conditions)
    n_conditions = np.empty(n_variables, dtype=int)
    for i_variable in range(n_variables):
        n_conditions[i_variable] = len(conditions[i_variable])
    return n_conditions



import numpy as np
from ccalafiore.combinations import n_conditions_to_combinations, conditions_to_combinations
from ccalafiore.array import samples_in_arr1_are_in_arr2


def merge_axes_old(
        array_input,
        axes_removing_input,
        axis_samples_input,
        dtype=None):

    # Input requirements:
    # 1) axes_removing_input cannot contain same values;
    # 2) variables_inserting_table_output cannot contain same values;
    # 3) shapes of axes_removing_input and variables_inserting_table_output must be equal
    # 4) axes_removing_input[a] != axis_samples_input != axis_variables_table_input

    # Import requirements:
    # 1) import numpy as np
    # 2) from ccalafiore.combinations import n_conditions_to_combinations
    # 3) from ccalafiore.array import samples_in_arr1_are_in_arr2
    # 4) from ccalafiore.array import advanced_indexing

    # from ccalafiore.array import samples_in_arr1_are_in_arr2, advanced_indexing

    # format axes_removing_input
    try:
        n_axes_removing_input = len(axes_removing_input)
        axes_removing_input = np.asarray(axes_removing_input, dtype=int)
    except TypeError:
        axes_removing_input = np.asarray([axes_removing_input], dtype=int)
        n_axes_removing_input = len(axes_removing_input)

    shape_array_input = np.asarray(array_input.shape)
    n_axes_array_input = len(shape_array_input)

    if axis_samples_input < 0:
        axis_samples_input += n_axes_array_input
    axes_removing_input[axes_removing_input < 0] += n_axes_array_input

    # check point 1
    if np.sum(axes_removing_input[0] == axes_removing_input) > 1:
        raise Exception('axes_removing_input cannot contain repeated values')

    # check point 4
    if np.sum(axes_removing_input == axis_samples_input) > 0:
        raise Exception('axes_removing_input[a] != axis_samples_input')

    axes_array_input = np.arange(n_axes_array_input)
    axes_other_array_input = axes_array_input[np.logical_not(samples_in_arr1_are_in_arr2(
        axes_array_input, np.append(axis_samples_input, axes_removing_input)))]

    # axes_other_array_input_inverted = axes_other_array_input[::-1]
    # n_axes_other_array_input = len(axes_other_array_input)

    axis_samples_output = axis_samples_input - np.sum(axes_removing_input < axis_samples_input)

    n_axes_array_output = n_axes_array_input - n_axes_removing_input
    axes_array_output = np.arange(n_axes_array_output)
    axes_other_output = axes_array_output[axes_array_output != axis_samples_output]

    shape_array_output = np.empty(n_axes_array_output, dtype=object)
    n_conditions_in_axes_removing_input = shape_array_input[axes_removing_input]
    shape_array_output[axis_samples_output] = (
            shape_array_input[axis_samples_input] * np.prod(n_conditions_in_axes_removing_input))
    shape_array_output[axes_other_output] = shape_array_input[axes_other_array_input]

    if dtype is None:
        dtype = array_input.dtype
    array_output = np.empty(shape_array_output, dtype=dtype)

    combinations_axes_removing_input = n_conditions_to_combinations(n_conditions_in_axes_removing_input)
    n_combinations_axes_removing_input = combinations_axes_removing_input.shape[0]

    indexes_combinations_removing_input = np.empty(2, dtype=object)
    indexes_combinations_removing_input[:] = slice(None)

    indexes_combinations_removing_output = np.empty(2, dtype=object)
    indexes_combinations_removing_output[:] = slice(None)

    axes_removing_input_inverted = np.sort(axes_removing_input)[::-1]

    # index_input = np.full(n_axes_array_input, slice(None), dtype=object)
    indexes_input = np.empty(n_axes_array_input, dtype=object)
    indexes_input[:] = slice(None)

    index_array_output = np.empty(n_axes_array_output, dtype=object)
    index_array_output[:] = slice(None)

    length_axis_flattering_input = shape_array_input[axis_samples_input]

    start_index_axis_samples_output = 0
    for c in range(n_combinations_axes_removing_input):

        stop_index_axis_samples_output = (c + 1) * length_axis_flattering_input
        index_array_output[axis_samples_output] = slice(
            start_index_axis_samples_output, stop_index_axis_samples_output)
        start_index_axis_samples_output = stop_index_axis_samples_output

        indexes_combinations_removing_input[0] = c
        indexes_input[axes_removing_input] = (
            combinations_axes_removing_input[tuple(indexes_combinations_removing_input)])

        array_output[tuple(index_array_output)] = array_input[tuple(indexes_input)]

    return array_output


def merge_axes(
        array_input,
        axes_merging_input,
        dtype=None):

    # Input requirements:
    # 1) axes_merging_input cannot contain same values;

    # Import requirements:
    # 1) import numpy as np
    # 2) from ccalafiore.combinations import n_conditions_to_combinations
    # 3) from ccalafiore.array import samples_in_arr1_are_in_arr2
    # 4) from ccalafiore.array import advanced_indexing

    # from ccalafiore.array import samples_in_arr1_are_in_arr2, advanced_indexing

    # format axes_merging_input
    try:
        n_axes_merging_input = len(axes_merging_input)
        axes_merging_input = np.asarray(axes_merging_input, dtype=int)
    except TypeError:
        axes_merging_input = np.asarray([axes_merging_input], dtype=int)
        n_axes_merging_input = 1

    shape_array_input = np.asarray(array_input.shape)
    n_axes_array_input = len(shape_array_input)
    axes_merging_input[axes_merging_input < 0] += n_axes_array_input

    # check point 1
    for a in axes_merging_input:
        if np.sum(a == axes_merging_input) > 1:
            raise ValueError('axes_merging_input cannot contain repeated values. {} is repeated.'. format(a))

    axes_array_input = np.arange(n_axes_array_input)
    axes_other_array_input = axes_array_input[np.logical_not(samples_in_arr1_are_in_arr2(
        axes_array_input, axes_merging_input))]

    axis_pooling_input = axes_merging_input[0]

    axes_removing_input = axes_merging_input[1:]
    n_axes_removing_input = axes_removing_input.size

    axis_pooling_output = axis_pooling_input - np.sum(axes_removing_input < axis_pooling_input)
    n_axes_array_output = n_axes_array_input - n_axes_removing_input
    axes_array_output = np.arange(n_axes_array_output)
    axes_other_output = axes_array_output[axes_array_output != axis_pooling_output]

    shape_array_output = np.empty(n_axes_array_output, dtype=object)
    n_conditions_in_axes_merging_input = shape_array_input[axes_merging_input]
    shape_array_output[axis_pooling_output] = np.prod(n_conditions_in_axes_merging_input)
    shape_array_output[axes_other_output] = shape_array_input[axes_other_array_input]

    if dtype is None:
        dtype = array_input.dtype
    array_output = np.empty(shape_array_output, dtype=dtype)

    n_conditions_in_axes_removing_input = shape_array_input[axes_removing_input]
    combinations_axes_removing_input = n_conditions_to_combinations(
        n_conditions_in_axes_removing_input, order_variables='lr')
    n_combinations_axes_removing_input = combinations_axes_removing_input.shape[0]

    indexes_combinations_removing_input = np.empty(2, dtype=object)
    indexes_combinations_removing_input[:] = slice(None)

    indexes_input = np.empty(n_axes_array_input, dtype=object)
    indexes_input[:] = slice(None)

    index_array_output = np.empty(n_axes_array_output, dtype=object)
    index_array_output[:] = slice(None)

    length_axis_merging_input_0 = shape_array_input[axis_pooling_input]

    start_index_axis_merged_output = 0
    for c in range(n_combinations_axes_removing_input):

        stop_index_axis_merged_output = (c + 1) * length_axis_merging_input_0
        index_array_output[axis_pooling_output] = slice(
            start_index_axis_merged_output, stop_index_axis_merged_output)
        start_index_axis_merged_output = stop_index_axis_merged_output

        indexes_combinations_removing_input[0] = c
        indexes_input[axes_removing_input] = (
            combinations_axes_removing_input[tuple(indexes_combinations_removing_input)])

        array_output[tuple(index_array_output)] = array_input[tuple(indexes_input)]

    return array_output


def merge_neighbouring_conditions(array_input, axis_1, axis_2, m=3, s=None):
    # axis_1 is (are) the axis (axes) that contains (contain) the merging conditions. it can be
    #        an integer or a list of integers. In case of list of integers, the list cannot contain repeated values.
    # axis_2 is the axis that contains the merging samples of a set of n neighbouring conditions in axis_1.
    #        it is always an integer.
    # m is (are) the number(s) of the merging conditions in axis_1. it is an integer is axis_1 is an integer or
    #        a list of integers if axis_1 is a list of integers. In case of list of int, its size has to be the
    #        same as the size of the axis_1.

    shape_array_input = np.asarray(array_input.shape)
    n_axes_array_input = shape_array_input.size

    # format axes_2
    if axis_2 < 0:
        axis_2 += n_axes_array_input

    # format axes_1
    try:
        n_axis_1 = len(axis_1)
        axis_1 = np.asarray(axis_1, dtype=int)
        axis_1[axis_1 < 0] += n_axes_array_input
        # check point 1
        if np.sum(axis_1[0] == axis_1) > 1:
            raise ValueError('axis_1 cannot contain repeated values')
    except TypeError:
        if axis_1 < 0:
            axis_1 += n_axes_array_input
        axis_1 = np.asarray([axis_1], dtype=int)
        n_axis_1 = 1

    axes_merging = np.append(axis_2, axis_1)

    # format m
    try:
        n_m = len(m)
        if n_m == n_axis_1:
            m = np.asarray(m, dtype=int)
        elif n_m == 1:
            m_tmp = m[0]
            m = np.empty(n_axis_1, dtype=int)
            m[:] = m_tmp
            n_s = n_axis_1
        else:
            # check point 3
            raise ValueError('sizes of axis_1 and m must be equal')
    except TypeError:
        m_tmp = m
        m = np.empty(n_axis_1, dtype=int)
        m[:] = m_tmp
        n_m = n_axis_1

    # format s
    if s is None:
        s = m
    else:
        try:
            n_s = len(s)
            if n_s == n_axis_1:
                s = np.asarray(s, dtype=int)
                for a in range(n_axis_1):
                    if s[a] is None:
                        s[a] = m[a]
            elif n_s == 1:
                s_tmp = s[0]
                if s_tmp is None:
                    s = m
                else:
                    s = np.empty(n_axis_1, dtype=int)
                    s[:] = s_tmp
                n_s = n_axis_1
            else:
                # check point 3
                raise ValueError('sizes of axis_1 and s must be equal')
        except TypeError:
            s_tmp = s
            s = np.empty(n_axis_1, dtype=int)
            s[:] = s_tmp
            n_s = n_axis_1

    r_1 = m // 2
    r_0 = r_1
    m_even = (m % 2) == 0
    if any(m_even):
        r_0[m_even] -= 1

    n_conditions_input = shape_array_input[axis_1]
    n_samples_per_condition_input = shape_array_input[axis_2]

    n_conditions_output = ((n_conditions_input - m) // s) + 1
    n_samples_per_m_conditions_output = n_samples_per_condition_input * np.prod(m)

    shape_array_output = shape_array_input
    shape_array_output[axis_1] = n_conditions_output
    shape_array_output[axis_2] = n_samples_per_m_conditions_output
    array_output = np.empty(shape_array_output, dtype=array_input.dtype)

    index_input = np.empty(n_axes_array_input, dtype=object)
    index_input[:] = slice(None)
    index_output = np.copy(index_input)

    center = np.empty(n_axis_1, dtype=object)
    for a in range(n_axis_1):
        center[a] = np.arange(r_0[a], n_conditions_input[a] - r_1[a], s[a])
    combinations_centers = conditions_to_combinations(center)
    n_combinations_centers = combinations_centers.shape[0]

    combinations_merged_conditions_output = n_conditions_to_combinations(n_conditions_output)
    for i in range(n_combinations_centers):
        for a in range(n_axis_1):
            index_input[axis_1[a]] = slice(combinations_centers[i, a] - r_0[a], combinations_centers[i, a] + r_1[a] + 1)
        index_output[axis_1] = combinations_merged_conditions_output[i]

        # array_output[tuple(index_output)] = merge_axes(array_input[tuple(index_input)], axis_1, axis_2)

        array_output[tuple(index_output)] = merge_axes(array_input[tuple(index_input)], axes_merging)

    return array_output



import numpy as np
from ccalafiore.combinations import make_combinations_of_conditions_as_unique_values


def passwords_in_txt_files(directory_txt_file, n_digits=3):

    # characters = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890`¦!"$%^&*()_-+={}[]:;\'@#|\\,<.>/?'
    # print(characters)
    # characters = ''.join(sorted(characters))
    # print(characters)

    # characters = '!"#$%&\'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}¦'
    characters = np.array(['0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
                           'A', 'B', 'C', 'D' ,'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R',
                           'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
                           'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r',
                           's', 't', 'u', 'v', 'w', 'x', 'y', 'z'])

    print(len(characters))
    characters = np.asanyarray(sorted(characters), dtype='<U1')
    print(characters)
    print(len(characters))
    conditions_characters = np.empty(n_digits, dtype=object)
    for i_digit in range(n_digits):
        conditions_characters[i_digit] = characters

    passwords = make_combinations_of_conditions_as_unique_values(conditions_characters, type_conditions_values='<U1')

    np.savetxt(directory_txt_file, passwords, fmt="%s", delimiter="")





import numpy as np
from ccalafiore.array import samples_in_arr1_are_in_arr2, advanced_indexing
# from scipy.stats import t
# from scipy.stats.stats import _ttest_finish
# from ccalafiore.combinations import n_conditions_to_combinations


def format_scores_for_paired_t_test(scores_raw, axis_independent_variable=-2, dtype=None):

    shape_scores_raw = np.asarray(scores_raw.shape)
    n_axes_scores_raw = len(shape_scores_raw)
    axis_independent_variable_raw = axis_independent_variable % n_axes_scores_raw

    scores_raw = np.expand_dims(scores_raw, axis=axis_independent_variable_raw)
    scores_raw = np.expand_dims(scores_raw, axis=axis_independent_variable_raw)
    axis_independent_variable_raw += 2

    n_conditions_iv_per_comparisons = 2

    shape_scores_raw = np.asarray(scores_raw.shape)
    n_axes_scores_raw = len(shape_scores_raw)
    n_conditions_iv = shape_scores_raw[axis_independent_variable_raw]

    axes_scores = np.arange(n_axes_scores_raw)
    axes_non_independent_variables_scores_raw = axes_scores[np.logical_not(samples_in_arr1_are_in_arr2(
        axes_scores, axis_independent_variable_raw))]

    index_0_raw = np.empty(n_axes_scores_raw, dtype=object)
    index_0_formatted = np.copy(index_0_raw)

    for a in axes_non_independent_variables_scores_raw:
        index_0_raw[a] = np.arange(shape_scores_raw[a])
    index_1_raw = np.copy(index_0_raw)

    axes_independent_variable_formatted = np.asarray([
        axis_independent_variable_raw - 2,
        axis_independent_variable_raw - 1,
        axis_independent_variable_raw], dtype=int)

    axes_non_independent_variable_formatted = axes_scores[np.logical_not(samples_in_arr1_are_in_arr2(
        axes_scores, axes_independent_variable_formatted))]

    index_0_formatted[axes_non_independent_variable_formatted] = index_0_raw[axes_non_independent_variable_formatted]
    index_0_formatted[axes_independent_variable_formatted[0]] = np.arange(2)
    index_1_formatted = np.copy(index_0_formatted)

    shape_scores_formatted = np.copy(shape_scores_raw)
    shape_scores_formatted[axes_independent_variable_formatted[0]] = n_conditions_iv_per_comparisons
    shape_scores_formatted[axes_independent_variable_formatted[1:]] = n_conditions_iv

    if dtype is None:
        dtype = scores_raw.dtype
    # scores_formatted = np.full(shape_scores_formatted, 0, dtype=dtype)
    scores_formatted = np.empty(shape_scores_formatted, dtype=dtype)

    for i_condition_0 in range(n_conditions_iv):

        index_0_raw[axis_independent_variable_raw] = i_condition_0
        index_0_raw_adv = advanced_indexing(index_0_raw)

        index_0_formatted[axes_independent_variable_formatted[1]] = i_condition_0
        index_1_formatted[axes_independent_variable_formatted[2]] = i_condition_0

        for i_condition_1 in range(i_condition_0, n_conditions_iv):

            scores_raw_0 = scores_raw[index_0_raw_adv]

            index_0_formatted[axes_independent_variable_formatted[2]] = i_condition_1
            index_0_formatted_adv = advanced_indexing(index_0_formatted)

            if i_condition_0 == i_condition_1:

                scores_formatted[index_0_formatted_adv] = np.append(
                    scores_raw_0, scores_raw_0, axis=axes_independent_variable_formatted[0])

            else:
                index_1_raw[axis_independent_variable_raw] = i_condition_1
                index_1_raw_adv = advanced_indexing(index_1_raw)

                scores_raw_1 = scores_raw[index_1_raw_adv]

                index_1_formatted[axes_independent_variable_formatted[1]] = i_condition_1
                index_1_formatted_adv = advanced_indexing(index_1_formatted)

                scores_formatted[index_0_formatted_adv] = np.append(
                    scores_raw_0, scores_raw_1, axis=axes_independent_variable_formatted[0])

                scores_formatted[index_1_formatted_adv] = np.append(
                    scores_raw_1, scores_raw_0, axis=axes_independent_variable_formatted[0])

    return scores_formatted


def scores_to_diff_of_scores(scores, axis_independent_variable=-2, dtype=None):

    shape_scores = np.asarray(scores.shape)
    n_axes_scores = len(shape_scores)
    axis_independent_variable %= n_axes_scores

    scores = np.expand_dims(scores, axis=axis_independent_variable + 1)
    shape_scores = np.asarray(scores.shape)
    n_axes_scores = len(shape_scores)
    n_conditions_iv = shape_scores[axis_independent_variable]

    axes_scores = np.arange(n_axes_scores)
    axes_non_independent_variables_scores = axes_scores[np.logical_not(samples_in_arr1_are_in_arr2(
        axes_scores, axis_independent_variable))]

    index_1 = np.empty(n_axes_scores, dtype=object)
    index_diff_1 = np.copy(index_1)

    for a in axes_non_independent_variables_scores:
        index_1[a] = np.arange(shape_scores[a])
    index_2 = np.copy(index_1)

    independent_variables_diff = np.asarray([axis_independent_variable, axis_independent_variable + 1], dtype=int)
    axes_non_independent_variables_diff = axes_scores[np.logical_not(samples_in_arr1_are_in_arr2(
        axes_scores, independent_variables_diff))]

    index_diff_1[axes_non_independent_variables_diff] = index_1[axes_non_independent_variables_diff]
    index_diff_2 = np.copy(index_diff_1)

    shape_diff = np.copy(shape_scores)
    shape_diff[independent_variables_diff] = n_conditions_iv
    if dtype is None:
        dtype = scores.dtype
    diff_of_scores = np.full(shape_diff, 0, dtype=dtype)

    for i_condition_1 in range(n_conditions_iv - 1):

        index_1[axis_independent_variable] = i_condition_1
        index_1_adv = advanced_indexing(index_1)
        index_diff_1[independent_variables_diff[0]] = i_condition_1
        index_diff_2[independent_variables_diff[1]] = i_condition_1

        for i_condition_2 in range(i_condition_1 + 1, n_conditions_iv):

            index_2[axis_independent_variable] = i_condition_2
            index_2_adv = advanced_indexing(index_2)

            index_diff_1[independent_variables_diff[1]] = i_condition_2
            index_diff_1_adv = advanced_indexing(index_diff_1)

            index_diff_2[independent_variables_diff[0]] = i_condition_2
            index_diff_2_adv = advanced_indexing(index_diff_2)

            diff_of_scores_1_2 = scores[index_1_adv] - scores[index_2_adv]
            diff_of_scores[index_diff_1_adv] = diff_of_scores_1_2
            diff_of_scores[index_diff_2_adv] = -diff_of_scores_1_2 + 0

    return diff_of_scores





import numpy as np


def rotate_indexes(array, axes_rotating, m):

    # This function rotates the indexes of "array" on the "axes" by "m" indexes.
    # It returns None.

    shape_array = np.asarray(array.shape, dtype=int)
    n_axes = shape_array.size

    # format axes
    try:
        n_axes_rotating = len(axes_rotating)
        axes_rotating = np.asarray(axes_rotating, dtype=int)
        axes_rotating[axes_rotating < 0] += n_axes
    except TypeError:
        if axes_rotating < 0:
            axes_rotating += n_axes
        axes_rotating = np.asarray([axes_rotating], dtype=int)
        n_axes_rotating = 1

    n_indexes = shape_array[axes_rotating]
    # format m
    try:
        n_m = len(m)
        m = np.asarray(m, dtype=int)

        if n_m != n_axes_rotating:
            if n_m == 1:
                m_tmp = m[0]
                m = np.empty(n_axes_rotating, dtype=int)
                m[:] = m_tmp
                n_m = n_axes_rotating
            else:
                raise ValueError('Shapes of axes_inserting and variables_table_adding_axes must be equal')

    except TypeError:
        m = np.asarray([m], dtype=int)
        n_m = 1
        if n_m != n_axes_rotating:
            m_tmp = m[0]
            m = np.empty(n_axes_rotating, dtype=int)
            m[:] = m_tmp
            n_m = n_axes_rotating

    m %= n_indexes

    index_in = np.empty(n_axes, dtype=object)

    for a in axes_rotating:
        index_in[:] = slice(None)
        index_out = np.copy(index_in)

        index_in[a] = slice(n_indexes[a] - m[a], n_indexes[a], 1)
        array_out_0_to_m = np.copy(array[tuple(index_in)])

        index_in[a] = slice(0, n_indexes[a] - m[a], 1)
        index_out[a] = slice(m[a], n_indexes[a], 1)
        array[tuple(index_out)] = array[tuple(index_in)]

        index_out[a] = slice(0, m[a], 1)
        array[tuple(index_out)] = array_out_0_to_m

    return None



"""
===========================
Creating annotated heatmaps
===========================

It is often desirable to show data which depends on two independent
variables as a color coded image plot. This is often referred to as a
heatmap. If the data is categorical, this would be called a categorical
heatmap.
Matplotlib's :meth:`imshow <matplotlib.axes.Axes.imshow>` function makes
production of such plots particularly easy.

The following examples show how to create a heatmap with annotations.
We will start with an easy example and expand it to be usable as a
universal function.
"""


##############################################################################
#
# A simple categorical heatmap
# ----------------------------
#
# We may start by defining some data. What we need is a 2D list or array
# which defines the data to color code. We then also need two lists or arrays
# of categories; of course the number of elements in those lists
# need to match the data along the respective axes.
# The heatmap itself is an :meth:`imshow <matplotlib.axes.Axes.imshow>` plot
# with the labels set to the categories we have.
# Note that it is important to set both, the tick locations
# (:meth:`set_xticks<matplotlib.axes.Axes.set_xticks>`) as well as the
# tick labels (:meth:`set_xticklabels<matplotlib.axes.Axes.set_xticklabels>`),
# otherwise they would become out of sync. The locations are just
# the ascending integer numbers, while the ticklabels are the labels to show.
# Finally we can label the data itself by creating a
# :class:`~matplotlib.text.Text` within each cell showing the value of
# that cell.


# import numpy as np
from matplotlib import pyplot as plt, ticker as ticker
from mpl_toolkits.axes_grid1 import make_axes_locatable
# sphinx_gallery_thumbnail_number = 2


#############################################################################
# Using the helper function code style
# ------------------------------------
#
# As discussed in the :ref:`Coding styles <coding_styles>`
# one might want to reuse such code to create some kind of heatmap
# for different input data and/or on different axes.
# We create a function that takes the data and the row and column labels as
# input, and allows arguments that are used to customize the plot
#
# Here, in addition to the above we also want to create a colorbar and
# position the labels above of the heatmap instead of below it.
# The annotations shall get different colors depending on a threshold
# for better contrast against the pixel color.
# Finally, we turn the surrounding axes spines off and create
# a grid of white lines to separate the cells.


def heatmap(data, cmap=None, ax=None,
            title=None, x_label=None, y_label=None,
            annot=False, fmt_annot='.2f', fontsize_annot=None, textcolors="black",
            xticklabels='auto', yticklabels='auto',
            fontsize_title=None, rotation_title=None,
            fontsize_x_label=None, rotation_x_label=None,
            fontsize_y_label=None, rotation_y_label=None,
            fontsize_xticklabels=None, rotation_xticklabels=None,
            fontsize_yticklabels=None, rotation_yticklabels=None,
            fontsize_cbar_ticklabels=None, cbarlabel="", fontsize_cbarlabel=None):
    """
    Create a heatmap from a numpy array and two lists of labels.

    Parameters
    ----------
    data
        A 2D numpy array of shape (N, M).
    y_labels
        A list or array of length N with the labels for the rows.
    x_labels
        A list or array of length M with the labels for the columns.
    ax
        A `matplotlib.axes.Axes` instance to which the heatmap is plotted.  If
        not provided, use current axes or create a new one.  Optional.
    cbar_kw
        A dictionary with arguments to `matplotlib.Figure.colorbar`.  Optional.
    cbarlabel
        The label for the colorbar.  Optional.
    **kwargs
        All other arguments are forwarded to `imshow`.
    """

    if ax is None:
        figures_existing = plt.get_fignums()
        n_figures_new = 1
        i = 0
        f = 0
        while f < n_figures_new:
            if i in figures_existing:
                pass
            else:
                id_figures = i
                f += 1
            i += 1
        if data.shape[1] > data.shape[0]:
            ratio_x = 1
            ratio_y = data.shape[0] / data.shape[1]
        elif data.shape[1] < data.shape[0]:
            ratio_x = data.shape[1] / data.shape[0]
            ratio_y = 1
        else:
            ratio_x = 1
            ratio_y = 1

        plt.figure(id_figures, figsize=(10 * ratio_x, 10 * ratio_y))
        ax = plt.gca()

    # Plot the heatmap
    im = ax.imshow(data, cmap=cmap)

    # Create colorbar
    # divider = make_axes_locatable(ax)
    # cax = divider.append_axes("right", size="5%", pad=0.05)
    cbar = plt.colorbar(im, ax=ax)
    ax.set_aspect('auto')
    cbar.ax.tick_params(labelsize=fontsize_cbar_ticklabels)
    cbar.ax.set_ylabel(cbarlabel, rotation=-90, va="bottom", fontsize=fontsize_cbarlabel)
    # if fontsize_cbar_ticklabels != 'auto':
    #     # fontsize_cbar_ticklabels = 8
    #     # use matplotlib.colorbar.Colorbar object
    #     cbar = ax.collections[0].colorbar
    #     # here set the labelsize
    #     cbar.ax.tick_params(labelsize=fontsize_cbar_ticklabels)

    plt.title(title, fontsize=fontsize_title, rotation=rotation_title)
    if x_label is not None:
        plt.xlabel(x_label, fontsize=fontsize_x_label, rotation=rotation_x_label)
    if y_label is not None:
        plt.ylabel(y_label, fontsize=fontsize_y_label, rotation=rotation_y_label)

    # # We want to show all ticks...
    ax.set_xticks(list(range(data.shape[1])))
    ax.set_yticks(list(range(data.shape[0])))
    if xticklabels != 'auto':
        # ... and label them with the respective list entries.
        ax.set_xticklabels(xticklabels, fontsize=fontsize_xticklabels, rotation=rotation_xticklabels)

    if yticklabels != 'auto':
        ax.set_yticklabels(yticklabels, fontsize=fontsize_yticklabels, rotation=rotation_yticklabels)


    if annot:
        """"
        A function to annotate a heatmap.
    
        Parameters
        ----------
        im
            The AxesImage to be labeled.
        data
            Data used to annotate.  If None, the image's data is used.  Optional.
        valfmt
            The format of the annotations inside the heatmap.  This should either
            use the string format method, e.g. "$ {x:.2f}", or be a
            `matplotlib.ticker.Formatter`.  Optional.
        textcolors
            A list or array of two color specifications.  The first is used for
            values below a threshold, the second for those above.  Optional.
        threshold
            Value in data units according to which the colors from textcolors are
            applied.  If None (the default) uses the middle of the colormap as
            separation.  Optional.
        **kwargs
            All other arguments are forwarded to each call to `text` used to create
            the text labels.
        """

        # Set default alignment to center, but allow it to be
        # overwritten by textkw.
        kw = dict(horizontalalignment="center",
                  verticalalignment="center",
                  color=textcolors,
                  fontsize=fontsize_annot)

        # Get the formatter in case a string is supplied
        # if isinstance(fmt_annot, str):
        #     fmt_annot = ticker.StrMethodFormatter(fmt_annot)

        # Loop over the data and create a `Text` for each "pixel".
        # Change the text's color depending on the data.
        # texts = []
        # print()
        for i in range(data.shape[0]):
            for j in range(data.shape[1]):
                # kw.update(color=textcolors[int(im.norm(data[i, j]) > threshold)])
                text = im.axes.text(j, i, fmt_annot.format( data[0, 0]),**kw)
                # texts.append(text)

        # return texts
    plt.tight_layout()
    plt.show()


